from tgt_grease.core import GreaseContainer
from tgt_grease.enterprise.Model import Scheduling
from tgt_grease.enterprise.Model import Scheduler
from tgt_grease.enterprise.Model import Deduplication
from bson.objectid import ObjectId
import datetime


class NodeMonitoring(object):
    """Monitors cluster nodes for unhealthy state

    Attributes:
        ioc (GreaseContainer): IoC Access
        centralScheduler (Scheduling): Central Scheduling Instance
        scheduler (Scheduler): Scheduling Model Instance

    """

    def __init__(self, ioc=None):
        if isinstance(ioc, GreaseContainer):
            self.ioc = ioc
        else:
            self.ioc = GreaseContainer()
        self.centralScheduler = Scheduling(self.ioc)
        self.scheduler = Scheduler(self.ioc)

    def monitor(self):
        """Monitoring process

        Returns:
            bool: If successful monitoring run occurred

        """
        servers = self.getServers()
        retVal = False
        self.ioc.getLogger().debug("Total servers to monitor [{0}]".format(len(servers)), trace=True)
        for server in servers:
            if self.serverAlive(server.get('_id')):
                retVal = True
                continue
            else:
                self.ioc.getLogger().warning("Server [{0}] preparing to be culled from pool".format(server.get('_id')))
                self.ioc.getLogger().warning("Server [{0}] preparing to be deactivated".format(server.get('_id')))
                if not self.deactivateServer(server.get('_id')):
                    self.ioc.getLogger().error(
                        "Failed deactivating server [{0}]".format(server.get('_id'))
                    )
                    retVal = False
                    break
                self.ioc.getLogger().warning(
                    "Server [{0}] preparing to reallocate detect jobs".format(server.get('_id'))
                )
                if not self.rescheduleDetectJobs(server.get('_id')):
                    self.ioc.getLogger().error(
                        "Failed rescheduling detect jobs [{0}]".format(server.get('_id'))
                    )
                    retVal = False
                    break
                self.ioc.getLogger().warning(
                    "Server [{0}] preparing to reallocate schedule jobs".format(server.get('_id'))
                )
                if not self.rescheduleScheduleJobs(server.get('_id')):
                    self.ioc.getLogger().error(
                        "Failed rescheduling detect jobs [{0}]".format(server.get('_id'))
                    )
                    retVal = False
                    break
                self.ioc.getLogger().warning(
                    "Server [{0}] preparing to reallocate jobs".format(server.get('_id'))
                )
                if not self.rescheduleJobs(server.get('_id')):
                    self.ioc.getLogger().error(
                        "Failed rescheduling detect jobs [{0}]".format(server.get('_id'))
                    )
                    retVal = False
                    break
        return retVal

    def scanComplete(self):
        """Enters a completed source so that this local server is alive next run

        This method is so that the server's 'heart' beats after each run. It will insert a completed SourceData document
        and increments the job counter in the JobServer Document

        Returns:
            None: Writes a MongoDB Document

        """
        self.ioc.getCollection('SourceData').insert_one({
            'grease_data': {
                'sourcing': {
                    'server': ObjectId(self.ioc.getConfig().NodeIdentity)
                },
                'detection': {
                    'server': ObjectId(self.ioc.getConfig().NodeIdentity),
                    'start': datetime.datetime.utcnow(),
                    'end': datetime.datetime.utcnow(),
                    'detection': {}
                },
                'scheduling': {
                    'server': ObjectId(self.ioc.getConfig().NodeIdentity),
                    'start': datetime.datetime.utcnow(),
                    'end': datetime.datetime.utcnow()
                },
                'execution': {
                    'server': ObjectId(self.ioc.getConfig().NodeIdentity),
                    'assignmentTime': datetime.datetime.utcnow(),
                    'completeTime': datetime.datetime.utcnow(),
                    'returnData': {},
                    'executionSuccess': True,
                    'commandSuccess': True,
                    'failures': 0
                }
            },
            'source': 'grease_internal_node_monitoring',
            'configuration': None,
            'data': {},
            'createTime': datetime.datetime.utcnow(),
            'expiry': Deduplication.generate_max_expiry_time(1)
        })
        server = self.ioc.getCollection('JobServer').find_one({'_id': ObjectId(self.ioc.getConfig().NodeIdentity)})
        if not server:
            self.ioc.getLogger().critical(
                "Failed to find server [{0}] after monitoring occurred!".format(self.ioc.getConfig().NodeIdentity)
            )
        self.ioc.getCollection('JobServer').update_one({
            '_id': ObjectId(self.ioc.getConfig().NodeIdentity)},
            {'$set': {'jobs': dict(server).get('jobs', 0) + 1}}
        )

    def getServers(self):
        """Returns the servers to be monitored this cycle

        Returns:
            list[dict]: List of servers

        """
        final = []
        servers = self.ioc.getCollection('JobServer').find({'active': True})
        for server in servers:
            final.append(dict(server))
        return final

    def serverAlive(self, serverId):
        """Checks to see if server is alive

        This method checks if the serverID exists in the collection and determines if it's execution number has
        changed recently. If it is a newly configured node it will be added to the monitoring collection

        Args:
            serverId (str): ObjectId of server

        Returns:
            bool: If server is alive

        """
        # Server Health Collection
        coll = self.ioc.getCollection('ServerHealth')
        Server = coll.find_one({'server': ObjectId(serverId)})
        if Server:
            # We have a server already in the system
            serverStats = self.ioc.getCollection('JobServer').find_one({'_id': ObjectId(serverId)})
            if serverStats:
                # compare previous results to see if there has been change
                if dict(Server).get('jobs', 0) < dict(serverStats).get('jobs', 0):
                    # Job Server Numbers have changed
                    coll.update_one(
                        {'_id': Server['_id']},
                        {
                            '$set': {
                                'jobs': dict(serverStats).get('jobs', 0),
                                'checkTime': datetime.datetime.utcnow()
                            }
                        }
                    )
                    self.ioc.getLogger().trace("JobServer [{0}] is alive".format(serverId), trace=True)
                    return True
                else:
                    if dict(Server).get('checkTime', datetime.datetime.utcnow()) < \
                            datetime.datetime.utcnow() - datetime.timedelta(minutes=10):
                        # server has aged out
                        self.ioc.getLogger().trace(
                            "JobServer [{0}] is not alive; Timestamp has not changed in ten minutes".format(serverId),
                            trace=True
                        )
                        return False
                    else:
                        # server is in a degraded state
                        self.ioc.getLogger().warning("JobServer [{0}] is degraded!".format(serverId), trace=True)
                        return True
            else:
                # Failed to find server in JobServer collection
                self.ioc.getLogger().error("JobServer not found during node monitoring! [{0}]".format(serverId))
                return False
        else:
            # we have a new server
            serverStats = self.ioc.getCollection('JobServer').find_one({'_id': ObjectId(serverId)})
            if serverStats:
                coll.insert_one(
                    {
                        'server': ObjectId(serverId),
                        'jobs': dict(serverStats).get('jobs', 0),
                        'checkTime': datetime.datetime.utcnow()
                    }
                )
                self.ioc.getLogger().info("New JobServer persisted in monitoring [{0}]".format(serverId))
                return True
            else:
                # Failed to find server in JobServer collection
                self.ioc.getLogger().error("New JobServer not found during node monitoring! [{0}]".format(serverId))
                return False

    def deactivateServer(self, serverId):
        """deactivates server from pool

        Args:
            serverId (str): ObjectId to deactivate

        Returns:
            bool: If deactivation is successful

        """
        if self.ioc.getCollection('JobServer').update_one(
                {'_id': ObjectId(serverId)},
                {
                    '$set': {
                        'active': False
                    }
                }
        ).modified_count < 1:
            self.ioc.getLogger().warning("Server [{0}] failed to be deactivated".format(serverId))
            return False
        else:
            self.ioc.getLogger().warning("Server [{0}] deactivated".format(serverId))
            return True

    def rescheduleDetectJobs(self, serverId):
        """Reschedules any detection jobs

        Args:
            serverId (str): Server ObjectId

        Returns:
            bool: rescheduling success

        """
        retval = True
        server = self.ioc.getCollection('JobServer').find_one({'_id': ObjectId(serverId)})
        if not server:
            self.ioc.getLogger().error(
                "Failed to load server details while trying to reschedule detection [{0}]".format(serverId)
            )
            return False
        for job in self.ioc.getCollection('SourceData').find(
            {
                'grease_data.detection.server': ObjectId(serverId),
                'grease_data.detection.start': None,
                'grease_data.detection.end': None
            }
        ):
            job = dict(job)
            if not self.centralScheduler.scheduleDetection(job.get('source'), job.get('configuration'), [job]):
                retval = False
                break
            else:
                self.ioc.getCollection('JobServer').update_one(
                    {'_id': ObjectId(serverId)},
                    {
                        '$set': {
                            'jobs': dict(server).get('jobs', 0) - 1
                        }
                    }
                )
        return retval

    def rescheduleScheduleJobs(self, serverId):
        """Reschedules any detection jobs

        Args:
            serverId (str): Server ObjectId

        Returns:
            bool: rescheduling success

        """
        retval = True
        server = self.ioc.getCollection('JobServer').find_one({'_id': ObjectId(serverId)})
        if not server:
            self.ioc.getLogger().error(
                "Failed to load server details while trying to reschedule schedules [{0}]".format(serverId)
            )
            return False
        for job in self.ioc.getCollection('SourceData').find(
            {
                'grease_data.scheduling.server': ObjectId(serverId),
                'grease_data.scheduling.start': None,
                'grease_data.scheduling.end': None
            }
        ):
            job = dict(job)
            if not self.centralScheduler.scheduleScheduling(job.get('_id')):
                retval = False
                break
            else:
                self.ioc.getCollection('JobServer').update_one(
                    {'_id': ObjectId(serverId)},
                    {
                        '$set': {
                            'jobs': dict(server).get('jobs', 0) - 1
                        }
                    }
                )
        return retval

    def rescheduleJobs(self, serverId):
        """Reschedules any detection jobs

        Args:
            serverId (str): Server ObjectId

        Returns:
            bool: rescheduling success

        """
        retval = True
        server = self.ioc.getCollection('JobServer').find_one({'_id': ObjectId(serverId)})
        if not server:
            self.ioc.getLogger().error(
                "Failed to load server details while trying to reschedule schedules [{0}]".format(serverId)
            )
            return False
        for job in self.ioc.getCollection('SourceData').find(
            {
                'grease_data.execution.server': ObjectId(serverId),
                'grease_data.execution.commandSuccess': False,
                'grease_data.execution.executionSuccess': False,
                'grease_data.execution.failures': {'$lt': 6}
            }
        ):
            job = dict(job)
            if not self.scheduler.schedule(job):
                retval = False
                break
            else:
                self.ioc.getCollection('JobServer').update_one(
                    {'_id': ObjectId(serverId)},
                    {
                        '$set': {
                            'jobs': dict(server).get('jobs', 0) - 1
                        }
                    }
                )
        return retval



from unittest import TestCase
from tgt_grease.core import Mongo
import pymongo


class TestConnectivity(TestCase):
    def test_mongo_connection(self):
        mongo = Mongo()
        self.assertTrue(isinstance(mongo.Client(), pymongo.MongoClient))
        mongo.Close()

    def test_mongo_query(self):
        mongo = Mongo()
        client = mongo.Client()
        collection = client.get_database('grease').get_collection('test')
        collection.insert_one({'test': 'var'})
        result = collection.find({'test': 'var'}).count()
        self.assertTrue(result == 1)
        collection.drop()
        del collection
        del client
        mongo.Close()
        del mongo



from unittest import TestCase
from tgt_grease.enterprise.Model import Deduplication
from tgt_grease.core import GreaseContainer
import datetime
import hashlib
import uuid
import time


class TestDeduplication(TestCase):

    def test_comparison(self):
        self.assertTrue(Deduplication.string_match_percentage("Hello", "Hallo") == 0.8)

    def test_generate_expiry_time(self):
        self.assertTrue(
            Deduplication.generate_expiry_time(12).hour == (datetime.datetime.utcnow() + datetime.timedelta(hours=12)).hour
        )

    def test_generate_max_expiry_time(self):
            print(str(Deduplication.generate_max_expiry_time(7)))
            print(str(datetime.datetime.utcnow() + datetime.timedelta(days=7)))
            self.assertTrue(
                Deduplication.generate_max_expiry_time(7).day == (datetime.datetime.utcnow() + datetime.timedelta(days=7)).day
            )

    def test_generate_hash(self):
        obj = {'test': 'var', 'test1': 5, 'test2': 7.89}
        self.assertEqual(
            Deduplication.generate_hash_from_obj(obj),
            hashlib.sha256(str(obj).encode('utf-8')).hexdigest()
        )

    def test_generate_hash_multi_str_type(self):
        obj = {'test': u'var', 'test1': 5, 'test2': 7.89, 'test3': 'ver'}
        self.assertEqual(
            Deduplication.generate_hash_from_obj(obj),
            hashlib.sha256(str(obj).encode('utf-8')).hexdigest()
        )

    def test_generate_hash_other_type(self):
        obj = 7
        self.assertEqual(
            Deduplication.generate_hash_from_obj(obj),
            hashlib.sha256(str(obj).encode('utf-8')).hexdigest()
        )
        obj = 'test'
        self.assertEqual(
            Deduplication.generate_hash_from_obj(obj),
            hashlib.sha256(str(obj).encode('utf-8')).hexdigest()
        )
        obj = u'test'
        self.assertEqual(
            Deduplication.generate_hash_from_obj(obj),
            hashlib.sha256(str(obj).encode('utf-8')).hexdigest()
        )
        obj = 7.8
        self.assertEqual(
            Deduplication.generate_hash_from_obj(obj),
            hashlib.sha256(str(obj).encode('utf-8')).hexdigest()
        )
        obj = ['test', 'var', 8, 8.43]
        self.assertEqual(
            Deduplication.generate_hash_from_obj(obj),
            hashlib.sha256(str(obj).encode('utf-8')).hexdigest()
        )

    def test_object_score_low_duplication(self):
        obj1 = {
            'field1': 'value',
            'field2': 'value1',
            'field3': 'value2',
            'field4': 'value3',
            'field5': 'value4'
        }
        obj2 = {
            'field1': str(uuid.uuid4()),
            'field2':  str(uuid.uuid4()),
            'field3':  str(uuid.uuid4()),
            'field4':  str(uuid.uuid4()),
            'field5':  str(uuid.uuid4())
        }
        ioc = GreaseContainer()
        parent1 = ioc.getCollection('test_scoring').insert_one({
            'expiry': Deduplication.generate_expiry_time(1),
            'max_expiry': Deduplication.generate_max_expiry_time(1),
            'type': 1,
            'score': 1,
            'source': 'test_source',
            'hash': Deduplication.generate_hash_from_obj(obj1)
        }).inserted_id
        score1 = Deduplication.object_field_score(
            'test_scoring',
            ioc,
            'test_source',
            'test_configuration',
            obj1,
            parent1,
            1,
            1
        )
        parent2 = ioc.getCollection('test_scoring').insert_one({
            'expiry': Deduplication.generate_expiry_time(1),
            'max_expiry': Deduplication.generate_max_expiry_time(1),
            'type': 1,
            'score': 1,
            'source': 'test_source',
            'hash': Deduplication.generate_hash_from_obj(obj2)
        }).inserted_id
        score2 = Deduplication.object_field_score(
            'test_scoring',
            ioc,
            'test_source',
            'test_configuration',
            obj2,
            parent2,
            1,
            1
        )
        print("++++++++++++++++++++++++++++++++++")
        print("score1: {0}".format(score1))
        print("score2: {0}".format(score2))
        print("++++++++++++++++++++++++++++++++++")
        self.assertTrue(score1 == 0.0)
        self.assertTrue(score2 <= 20.0)
        ioc.getCollection('test_scoring').drop()
        time.sleep(1.5)

    def test_object_score_medium_duplication(self):
        obj1 = {
            'field1': 'value',
            'field2': 'value1',
            'field3': 'value2',
            'field4': 'value3',
            'field5': 'value4'
        }
        obj2 = {
            'field1': str(uuid.uuid4()),
            'field2': 'value1',
            'field3': 'value2',
            'field4':  str(uuid.uuid4()),
            'field5':  str(uuid.uuid4())
        }
        ioc = GreaseContainer()
        parent1 = ioc.getCollection('test_scoring').insert_one({
            'expiry': Deduplication.generate_expiry_time(1),
            'max_expiry': Deduplication.generate_max_expiry_time(1),
            'type': 1,
            'score': 1,
            'source': 'test_source',
            'hash': Deduplication.generate_hash_from_obj(obj1)
        }).inserted_id
        score1 = Deduplication.object_field_score(
            'test_scoring',
            ioc,
            'test_source',
            'test_configuration',
            obj1,
            parent1,
            1,
            1
        )
        parent2 = ioc.getCollection('test_scoring').insert_one({
            'expiry': Deduplication.generate_expiry_time(1),
            'max_expiry': Deduplication.generate_max_expiry_time(1),
            'type': 1,
            'score': 1,
            'source': 'test_source',
            'hash': Deduplication.generate_hash_from_obj(obj2)
        }).inserted_id
        score2 = Deduplication.object_field_score(
            'test_scoring',
            ioc,
            'test_source',
            'test_configuration',
            obj2,
            parent2,
            1,
            1
        )
        print("++++++++++++++++++++++++++++++++++")
        print("score1: {0}".format(score1))
        print("score2: {0}".format(score2))
        print("++++++++++++++++++++++++++++++++++")
        self.assertTrue(score1 == 0.0)
        self.assertTrue(score2 <= 50.0)
        ioc.getCollection('test_scoring').drop()
        time.sleep(1.5)

    def test_object_score_high_duplication(self):
        obj1 = {
            'field1': 'value',
            'field2': 'value1',
            'field3': 'value2',
            'field4': 'value3',
            'field5': 'value4'
        }
        obj2 = {
            'field1': str(uuid.uuid4()),
            'field2': 'value1',
            'field3': 'value2',
            'field4': 'value3',
            'field5': 'value4'
        }
        ioc = GreaseContainer()
        parent1 = ioc.getCollection('test_scoring').insert_one({
            'expiry': Deduplication.generate_expiry_time(1),
            'max_expiry': Deduplication.generate_max_expiry_time(1),
            'type': 1,
            'score': 1,
            'source': 'test_source',
            'hash': Deduplication.generate_hash_from_obj(obj1)
        }).inserted_id
        score1 = Deduplication.object_field_score(
            'test_scoring',
            ioc,
            'test_source',
            'test_configuration',
            obj1,
            parent1,
            1,
            1
        )
        parent2 = ioc.getCollection('test_scoring').insert_one({
            'expiry': Deduplication.generate_expiry_time(1),
            'max_expiry': Deduplication.generate_max_expiry_time(1),
            'type': 1,
            'score': 1,
            'source': 'test_source',
            'hash': Deduplication.generate_hash_from_obj(obj2)
        }).inserted_id
        score2 = Deduplication.object_field_score(
            'test_scoring',
            ioc,
            'test_source',
            'test_configuration',
            obj2,
            parent2,
            1,
            1
        )
        print("++++++++++++++++++++++++++++++++++")
        print("score1: {0}".format(score1))
        print("score2: {0}".format(score2))
        print("++++++++++++++++++++++++++++++++++")
        self.assertTrue(score1 == 0.0)
        self.assertTrue(score2 >= 80.0)
        ioc.getCollection('test_scoring').drop()
        time.sleep(1.5)

    def test_object_score_maximum_duplication(self):
        obj1 = {
            'field1': 'value',
            'field2': 'value1',
            'field3': 'value2',
            'field4': 'value3',
            'field5': 'value4'
        }
        obj2 = {
            'field1': 'value',
            'field2': 'value1',
            'field3': 'value2',
            'field4': 'value3',
            'field5': 'value4'
        }
        ioc = GreaseContainer()
        parent1 = ioc.getCollection('test_scoring').insert_one({
            'expiry': Deduplication.generate_expiry_time(1),
            'max_expiry': Deduplication.generate_max_expiry_time(1),
            'type': 1,
            'score': 1,
            'source': 'test_source',
            'hash': Deduplication.generate_hash_from_obj(obj1)
        }).inserted_id
        score1 = Deduplication.object_field_score(
            'test_scoring',
            ioc,
            'test_source',
            'test_configuration',
            obj1,
            parent1,
            1,
            1
        )
        parent2 = ioc.getCollection('test_scoring').insert_one({
            'expiry': Deduplication.generate_expiry_time(1),
            'max_expiry': Deduplication.generate_max_expiry_time(1),
            'type': 1,
            'score': 1,
            'source': 'test_source',
            'hash': Deduplication.generate_hash_from_obj(obj2)
        }).inserted_id
        score2 = Deduplication.object_field_score(
            'test_scoring',
            ioc,
            'test_source',
            'test_configuration',
            obj2,
            parent2,
            1,
            1
        )
        print("++++++++++++++++++++++++++++++++++")
        print("score1: {0}".format(score1))
        print("score2: {0}".format(score2))
        print("++++++++++++++++++++++++++++++++++")
        self.assertTrue(score1 == 0.0)
        self.assertTrue(score2 == 100.0)
        ioc.getCollection('test_scoring').drop()
        time.sleep(1.5)

    def test_deduplicate_object(self):
        ioc = GreaseContainer()
        ioc.getConfig().set('verbose', True, 'Logging')
        obj = [
            {
                'field': 'var',
                'field1': 'var1',
                'field2': 'var2',
                'field3': 'var3',
                'field4': 'var4',
                'field5': 'var5',
            },
            {
                'field': 'var',
                'field1': 'var1',
                'field2': 'var2',
                'field3': 'var3',
                'field4': 'var4',
                'field5': 'var5',
            },
            {
                'field': str(uuid.uuid4()),
                'field1': 'var1',
                'field2': str(uuid.uuid4()),
                'field3': str(uuid.uuid4()),
                'field4': 'var4',
                'field5': str(uuid.uuid4()),
            }
        ]
        finalObj = []
        Deduplication.deduplicate_object(
            ioc,
            obj[0],
            1,
            1,
            40.0,
            'test_source',
            'test_configuration',
            finalObj,
            'test_source'
        )
        self.assertEqual(len(finalObj), 1)
        Deduplication.deduplicate_object(
            ioc,
            obj[1],
            1,
            1,
            40.0,
            'test_source',
            'test_configuration',
            finalObj,
            'test_source'
        )
        self.assertEqual(len(finalObj), 1)
        Deduplication.deduplicate_object(
            ioc,
            obj[2],
            1,
            1,
            40.0,
            'test_source',
            'test_configuration',
            finalObj,
            'test_source'
        )
        self.assertGreaterEqual(len(finalObj), 1)
        ioc.getConfig().set('verbose', False, 'Logging')
        ioc.getCollection('test_source').drop()
        time.sleep(1.5)

    def test_deduplication(self):
        ioc = GreaseContainer()
        dedup = Deduplication(ioc)
        ioc.getConfig().set('verbose', True, 'Logging')
        obj = [
            {
                'field': 'var',
                'field1': 'var1',
                'field2': 'var2',
                'field3': 'var3',
                'field4': 'var4',
                'field5': 'var5',
            },
            {
                'field': 'var',
                'field1': 'var1',
                'field2': 'var2',
                'field3': 'var3',
                'field4': 'var4',
                'field5': 'var5',
            },
            {
                'field': 'var',
                'field1': 'var1',
                'field2': 'var2',
                'field3': 'var3',
                'field4': 'var4',
                'field5': 'var5',
            },
            {
                'field': 'var',
                'field1': 'var1',
                'field2': 'var2',
                'field3': 'var3',
                'field4': 'var4',
                'field5': 'var5',
            },
            {
                'field': str(uuid.uuid4()),
                'field1': str(uuid.uuid4()),
                'field2': str(uuid.uuid4()),
                'field3': str(uuid.uuid4()),
                'field4': str(uuid.uuid4()),
                'field5': str(uuid.uuid4())
            },
            {
                'field': 'var',
                'field1': 'var1',
                'field2': 'var2',
                'field3': 'var3',
                'field4': 'var4',
                'field5': 'var5',
            },
            {
                'field': 'var',
                'field1': 'var1',
                'field2': 'var2',
                'field3': 'var3',
                'field4': 'var4',
                'field5': 'var5',
            },
            {
                'field': 'var',
                'field1': 'var1',
                'field2': 'var2',
                'field3': 'var3',
                'field4': 'var4',
                'field5': 'var5',
            },
            {
                'field': str(uuid.uuid4()),
                'field1': str(uuid.uuid4()),
                'field2': str(uuid.uuid4()),
                'field3': str(uuid.uuid4()),
                'field4': str(uuid.uuid4()),
                'field5': str(uuid.uuid4())
            },
            {
                'field': str(uuid.uuid4()),
                'field1': str(uuid.uuid4()),
                'field2': str(uuid.uuid4()),
                'field3': str(uuid.uuid4()),
                'field4': str(uuid.uuid4()),
                'field5': str(uuid.uuid4())
            },
            {
                'field': str(uuid.uuid4()),
                'field1': str(uuid.uuid4()),
                'field2': str(uuid.uuid4()),
                'field3': str(uuid.uuid4()),
                'field4': str(uuid.uuid4()),
                'field5': str(uuid.uuid4())
            }
        ]
        finalObj = dedup.Deduplicate(
            obj,
            'test_source',
            'test_configuration',
            85.0,
            1,
            1,
            'test_source'
        )
        self.assertGreaterEqual(len(finalObj), 4)
        ioc.getConfig().set('verbose', False, 'Logging')
        ioc.getCollection('test_source').drop()
        time.sleep(1.5)



from tgt_grease.core import GreaseContainer
from bson.objectid import ObjectId
from .DeDuplication import Deduplication
import pymongo
import datetime


class Scheduling(object):
    """Central scheduling class for GREASE

    This class routes data to nodes within GREASE

    Attributes:
        ioc (GreaseContainer): IoC access for DeDuplication

    """

    def __init__(self, ioc=None):
        if isinstance(ioc, GreaseContainer):
            self.ioc = ioc
        else:
            self.ioc = GreaseContainer()
        self.ioc.ensureRegistration()

    def scheduleDetection(self, source, configName, data):
        """Schedule a Source Parse to detection

        This method will take a list of single dimension dictionaries and schedule them for detection

        Args:
            source (str): Name of the source
            configName (str): Configuration Data was sourced from
            data (list[dict]): Data to be scheduled for detection

        Returns:
            bool: Scheduling success

        """
        if len(data) is 0 or not isinstance(data, list):
            self.ioc.getLogger().trace(
                "Data provided empty or is not type list type: [{0}] len: [{1}]".format(str(type(data)), len(data)),
                trace=True
            )
            return False
        self.ioc.getLogger().trace("Preparing to schedule [{0}] source objects".format(len(data)), trace=True)
        sourceCollect = self.ioc.getCollection('SourceData')
        jServerCollect = self.ioc.getCollection('JobServer')
        # begin scheduling loop of each block
        for elem in data:
            if not isinstance(elem, dict):
                self.ioc.getLogger().warning(
                    "Element from data not of type dict! Got [{0}] DROPPED".format(str(type(elem))),
                    notify=False
                )
                continue
            server, jobCount = self.determineDetectionServer()
            if server:
                sourceCollect.insert_one({
                    'grease_data': {
                        'sourcing': {
                            'server': ObjectId(self.ioc.getConfig().NodeIdentity)
                        },
                        'detection': {
                            'server': ObjectId(server),
                            'start': None,
                            'end': None,
                            'detection': {}
                        },
                        'scheduling': {
                            'server': None,
                            'start': None,
                            'end': None
                        },
                        'execution': {
                            'server': None,
                            'assignmentTime': None,
                            'completeTime': None,
                            'returnData': {},
                            'executionSuccess': False,
                            'commandSuccess': False,
                            'failures': 0
                        }
                    },
                    'source': str(source).encode('utf-8'),
                    'configuration': str(configName).encode('utf-8'),
                    'data': elem,
                    'createTime': datetime.datetime.utcnow(),
                    'expiry': Deduplication.generate_max_expiry_time(1)
                })
                jServerCollect.update_one({
                    '_id': ObjectId(server)},
                    {'$set': {'jobs': int(jobCount) + 1}}
                )
            else:
                self.ioc.getLogger().warning(
                    "Failed to find detection server for data object from source [{0}]; DROPPED".format(source),
                    notify=False
                )
                self.ioc.getLogger().warning(
                    "Detection scheduling failed. Could not find detection server",
                    notify=False
                )
                return False
        return True

    def scheduleScheduling(self, objectId):
        """Schedule a source for job scheduling

        This method schedules a source for job scheduling

        Args:
            objectId (str): MongoDB ObjectId to schedule

        Returns:
            bool: If scheduling was successful

        """
        server, jobCount = self.determineSchedulingServer()
        if not server:
            self.ioc.getLogger().error("Failed to find scheduling server", notify=False)
            return False
        self.ioc.getCollection('SourceData').update_one(
            {'_id': ObjectId(objectId)},
            {
                '$set': {
                    'grease_data.scheduling.server': ObjectId(server),
                    'grease_data.scheduling.start': None,
                    'grease_data.scheduling.end': None
                }
            }
        )
        self.ioc.getCollection('SourceData').update_one({
            '_id': ObjectId(server)},
            {'$set': {'jobs': int(jobCount) + 1}}
        )
        return True

    def determineDetectionServer(self):
        """Determines detection server to use

        Finds the detection server available for a new detection job

        Returns:
            tuple: MongoDB Object ID of server & current job count

        """
        result = self.ioc.getCollection('JobServer').find({
            'active': True,
            'prototypes': 'detect'
        }).sort('jobs', pymongo.ASCENDING).limit(1)
        if result.count():
            return str(result[0]['_id']), int(result[0]['jobs'])
        else:
            return "", 0

    def determineSchedulingServer(self):
        """Determines scheduling server to use

        Finds the scheduling server available for a new scheduling job

        Returns:
            tuple: MongoDB Object ID of server & current job count

        """
        result = self.ioc.getCollection('JobServer').find({
            'active': True,
            'prototypes': 'schedule'
        }).sort('jobs', pymongo.DESCENDING).limit(1)
        if result.count():
            return str(result[0]['_id']), int(result[0]['jobs'])
        else:
            return "", 0

    def determineExecutionServer(self, role):
        """Determines execution server to use

        Finds the execution server available for a new execution job

        Returns:
            str: MongoDB Object ID of server; if one cannot be found then string will be empty

        """
        result = self.ioc.getCollection('JobServer').find({
            'active': True,
            'roles': str(role)
        }).sort('jobs', pymongo.DESCENDING).limit(1)
        if result.count():
            return str(result[0]['_id']), int(result[0]['jobs'])
        else:
            return "", 0



import json
from time import time
import threading
from multiprocessing import Pipe
import kafka
from kafka import KafkaConsumer
from tgt_grease.core import GreaseContainer
from tgt_grease.enterprise.Model.CentralScheduling import Scheduling
from .Configuration import PrototypeConfig

MIN_BACKLOG = 50     # If the Kafka message backlog falls below this number, we will kill a consumer
MAX_BACKLOG = 200    # If the Kafka message backlog rises above this number, we will make a consumer
SLEEP_TIME = 5       # Sleep this many seconds after creating or deleting a consumer.
MAX_CONSUMERS = 32   # We wont create more than this number of consumers for any config

class KafkaSource(object):
    """Kafka class for sourcing Kafka messages

    This Model will create and dynamically scale the number of Kafka consumers for the topics
    in the Config, and then sends the parsed messages (containing only the keys/values specified
    in the Config) to Scheduling.

    This Model is designed around the Configs. Each Config gets its own config_manager thread,
    which means Configs also get their own dedicated consumer. It was designed so that any
    "magic numbers" (such as MIN_BACKLOG, MAX_CONSUMERS, etc.) are overwriteable in the Config,
    with the exception of SLEEP_TIME, which can be constant accross Configs.

    Currently, the class only supports Kafka topics which contain JSON, however this functionality
    can easily be expanded on inside of the parse_message method.

    Attributes:
        ioc (GreaseContainer): IOC for scanning
        conf (PrototypeConfig): Prototype configuration instance
        configs (List[dict]): List of Kafka Configs

    Note:
        Currently, only json messages can be decoded from kafka topics

    """
    def __init__(self, ioc=None):
        if ioc and isinstance(ioc, GreaseContainer):
            self.ioc = ioc
        else:
            self.ioc = GreaseContainer()
        self.conf = PrototypeConfig(self.ioc)
        self.configs = []

    def run(self, config=None):
        """This will load all Kafka configs (unless a specific one is provided) and spin up consumer
        threads for all of them.

        It should never return anything unless something goes wrong with Kafka consumption.

        Creates a thread for each Kafka config to begin parsing messages. This parent thread then
        monitors its children, and prunes dead threads. Once all are dead, we return False.

        Note:
            If a configuration is set then *only* that configuration is parsed. If both are provided then the configuration will *only* be parsed if it is of the source provided.

        Args:
            config (dict): If set will only parse the specified config

        Returns:
            bool: False if an error occurs, else never returns

        """
        if config:
            self.configs = [config]
        else:
            self.configs = self.get_configs()

        if not self.validate_configs(self.configs):
            self.ioc.getLogger().error("One or more Kafka Configs are invalid, stopping.")
            return False

        threads = []
        for conf in self.configs:
            threads.append(self.create_consumer_manager_thread(conf))

        while threads:
            threads = list(filter(lambda x: x.is_alive(), threads))

        self.ioc.getLogger().critical("All Kafka consumer managers have died, stopping.")
        return False

    def create_consumer_manager_thread(self, config):
        """Creates and returns a thread running a consumer_manager

        Args:
            config (dict): Configuration for a Kafka Model

        Returns:
            threading.Thread: The thread running consumer_manager

        """
        KafkaSource.sleep(SLEEP_TIME)
        thread = threading.Thread(target=KafkaSource.consumer_manager, args=(self.ioc, config,))
        thread.daemon = False
        thread.start()
        self.ioc.getLogger().info("Kafka consumer manager thread started for config: {0}".format(config.get("name")))
        return thread

    @staticmethod
    def consumer_manager(ioc, config):
        """Creates and reallocates consumer threads within the same consumer group for a single config

        Args:
            ioc (GreaseContainer): Used for logging since we can't use self in threads
            config (dict): Configuration for a Kafka Model

        Returns:
            bool: False if all consumers are stopped

        """
        monitor_consumer = KafkaSource.create_consumer(ioc, config)
        threads = [KafkaSource.create_consumer_thread(ioc, config)]

        while threads:
            KafkaSource.reallocate_consumers(ioc, config, monitor_consumer, threads)
            threads = list(filter(lambda x: x[0].is_alive(), threads))

        return False

    @staticmethod
    def create_consumer_thread(ioc, config):
        """Creates a consumer thread, pipe pair for a given config

        Args:
            ioc (GreaseContainer): Used for logging since we can't use self in threads
            config (dict): Configuration for a Kafka Model

        Returns:
            threading.Thread: The Thread running the Kafka consumer
            multiprocessing.Pipe: The parent end of the Pipe used to send a kill signal to the consumer thread

        """
        parent_conn, child_conn = Pipe()
        thread = threading.Thread(target=KafkaSource.consume, args=(ioc, config, child_conn,))
        thread.daemon = True
        thread.start()
        ioc.getLogger().info("Kafka consumer thread started for config: {0}".format(config.get("name")))
        return thread, parent_conn

    @staticmethod
    def consume(ioc, config, pipe):
        """The Kafka consumer in charge of parsing messages according to the config, then sends the parsed dict to Scheduling

        Args:
            ioc (GreaseContainer): Used for logging since we can't use self in threads
            config (dict): Configuration for a Kafka Model
            pipe (multiprocessing.Pipe): Child end of the pipe used to receive signals from parent thread

        Returns:
            bool: False if kill signal is received

        """
        consumer = KafkaSource.create_consumer(ioc, config)

        for msg in consumer:
            if pipe.poll():    # If the parent pipe sends a signal
                ioc.getLogger().trace("Kill signal received, stopping", trace=True)
                return False
            message_dict = KafkaSource.parse_message(ioc, config, msg)
            if message_dict:
                KafkaSource.send_to_scheduling(ioc, config, message_dict)

        return False

    @staticmethod
    def sleep(sleep_sec):
        """Thread safe sleep function that waits sleep_sec seconds without affecting child threads

        Args:
            sleep_sec (int): Number of seconds to idle

        """
        wake_time = time() + sleep_sec
        while time() < wake_time:
            continue

    @staticmethod
    def create_consumer(ioc, config):
        """Creates a KafkaConsumer object from the params in config

        Args:
            ioc (GreaseContainer): Used for logging since we can't use self in threads
            config (dict): Configuration for a Kafka Model

        Returns:
            kafka.KafkaConsumer: KafkaConsumer object initialized with params from config

        """
        consumer = None
        while not consumer:
            try:
                consumer = KafkaConsumer(
                    group_id=config.get('name'),
                    *config.get('topics'),
                    **{'bootstrap_servers': ",".join(config.get('servers'))}
                )
            except kafka.errors.NoBrokersAvailable:
                ioc.getLogger().error("No Kafka brokers available for config: {0}, retrying.".format(config.get('name')))
                KafkaSource.sleep(SLEEP_TIME)

        ioc.getLogger().info("Kafka consumer created under group_id: {0}".format(config.get('name')))
        KafkaSource.sleep(SLEEP_TIME)   # Gives the consumer time to initialize
        return consumer

    @staticmethod
    def parse_message(ioc, config, message):
        """Parses a message from Kafka according to the config

        Note:
            transform_message extracts only the keys/values from the message as specified in the config. By default, we split the keys by "." - so if you wanted to access the value stored at message[a][b][c], your config would contain the key "a.b.c". You can overwrite the "." key splitter explicitly in your Config. These values will be written to their respective alias also specified in the config. 

        Args:
            ioc (GreaseContainer): Used for logging since we can't use self in threads
            config (dict): Configuration for a Kafka model
            message (kafka.ConsumerRecord): Individual message received from Kafka topic

        Returns:
            dict: A flat dictionary containing only the keys/values from the message as specified in the config

        """
        try:
            message = json.loads(message.value, strict=False)
            ioc.getLogger().trace("Message successfully loaded", trace=True)
        except ValueError:
            ioc.getLogger().trace("Failed to unload message", trace=True)
            return {}

        final = {}
        for key, alias in config.get("key_aliases", {}).items():
            pointer = message
            for sub_key in key.split(config.get("key_sep", ".")):
                if not isinstance(pointer, dict) or sub_key not in pointer:
                    ioc.getLogger().trace("Subkey: {0} missing from message".format(sub_key), trace=True)
                    return {}
                pointer = pointer[sub_key]
            final[alias] = pointer

        ioc.getLogger().trace("Message succesfully parsed", trace=True)
        return final

    @staticmethod
    def reallocate_consumers(ioc, config, monitor_consumer, threads):
        """Determines whether to create or kill a consumer based on current message backlog, then performs that action

        Args:
            ioc (GreaseContainer): Used for logging since we can't use self in threads
            config (dict): Configuration for a Kafka model
            monitor_consumer (kafka.KafkaConsumer): KafkaConsumer used solely for measuring message backlog
            threads (list[(threading.Thread, multiprocessing.Pipe)]): List of current consumer thread/pipe pairs

        Returns:
            int: Number of threads created (Negative value if a thread was killed)
        """
        min_backlog = config.get("min_backlog", MIN_BACKLOG)
        max_backlog = config.get("max_backlog", MAX_BACKLOG)
        max_consumers = config.get("max_consumers", MAX_CONSUMERS)

        backlog1 = KafkaSource.get_backlog(ioc, monitor_consumer)
        KafkaSource.sleep(SLEEP_TIME) # We want to wait before checking again in case there is a message spike
        backlog2 = KafkaSource.get_backlog(ioc, monitor_consumer)

        if backlog1 > max_backlog and backlog2 > max_backlog and len(threads) < max_consumers:
            threads.append(KafkaSource.create_consumer_thread(ioc, config))
            ioc.getLogger().info("Backlog max reached, spawning a new consumer for {0}".format(config.get('name')), verbose=True)
            return 1
        elif backlog1 <= min_backlog and backlog2 <= min_backlog and len(threads) > 1:
            KafkaSource.kill_consumer_thread(ioc, threads[0])
            ioc.getLogger().info("Backlog min reached, killing a consumer for {0}".format(config.get('name')), verbose=True)
            return -1
        ioc.getLogger().info("No reallocation needed for {0}".format(config.get('name')))
        return 0

    @staticmethod
    def kill_consumer_thread(ioc, thread_tup):
        """Sends a kill signal to the thread's pipe

        Note:
            Despite being from the multiprocessing library, Pipes are thread safe in this implementation as we don't share the same
            end of the Pipe to more than one thread. From the multiprocessing documentation:

                The two connection objects returned by Pipe() represent the two ends of the pipe. Each connection object has
                send() and recv() methods (among others). Note that data in a pipe may become corrupted if two threads
                (or threads) try to read from or write to the same end of the pipe at the same time. Of course there is no
                risk of corruption from threads using different ends of the pipe at the same time.

        Args:
            ioc (GreaseContainer): Used for logging since we can't use self in threads
            thread_tup ((threading.Thread, multiprocessing.Pipe)): Thread/Pipe tuple to be killed

        """
        thread_tup[1].send("STOP")
        ioc.getLogger().trace("Kill signal sent to consumer thread", trace=True)
        KafkaSource.sleep(SLEEP_TIME) # Give consumer a chance to finish its current message

    @staticmethod
    def get_backlog(ioc, consumer):
        """Gets the current message backlog for a given consumer

        Args:
            ioc (GreaseContainer): Used for logging since we can't use self in threads
            consumer (kafka.KafkaConsumer): The consumer used to poll backlog offsets

        Returns:
            float: the average number of messages accross all partitions in the backlog. -1 if there is an error and excess consumers should be killed

        """
        if not consumer.assignment():
            ioc.getLogger().trace("Assigning consumer to topic", trace=True)
            consumer.poll() # We need to poll the topic to actually get assigned

        partitions = consumer.assignment()
        if not partitions:
            ioc.getLogger().error("No partitions found for kafka consumer")
            return -1.

        try:
            current_offsets = [consumer.position(part) for part in partitions]
            end_offsets = list(consumer.end_offsets(partitions).values())
        except kafka.errors.KafkaTimeoutError:
            ioc.getLogger().error("KafkaTimeout during backlog check")
            return -1.
        except kafka.errors.UnsupportedVersionError:
            ioc.getLogger().error("This version of kafka does not support backlog lookups")
            return -1.

        if not current_offsets or not end_offsets or len(current_offsets) != len(end_offsets):
            ioc.getLogger().error("Backlog check failed for kafka consumer - invalid offsets")
            return -1.

        return float(sum(end_offsets) - sum(current_offsets)) / len(partitions)

    @staticmethod
    def send_to_scheduling(ioc, config, message):
        """Sends a parsed message dictionary to scheduling

        Args:
            ioc (GreaseContainer): Used for logging since we can't use self in threads
            config (dict): Configuration for a Kafka model
            message (dict): Individual parsed message received from Kafka topic

        Returns:
            bool: True if scheduling is successful

        """
        scheduler = Scheduling(ioc)
        if not message:
            return False
        if scheduler.scheduleDetection(config.get('source'), config.get('name'), message):
            ioc.getLogger().trace(
                "Data scheduled for detection from source [{0}]".format(config.get('source')),
                trace=True
            )
            return True
        else:
            ioc.getLogger().error("Scheduling failed for kafka source document!", notify=False)
            return False

    def get_configs(self):
        """Gets all Configs with the source 'kafka'

        Returns:
            list[dict]: A list of all kafka config dicts

        """
        self.ioc.getLogger().info("Kafka configs loaded")
        return self.conf.get_source('kafka')

    def validate_configs(self, configs):
        """Checks if configs all have the required keys and that there are no duplicate aliases

        Example Config::   
            {
                "name": "kafka_config",
                "source": "kafka",
                "key_aliases": {
                    "a*b*c": "abc_key",
                    "a*b*d": "abd_key"
                },
                "key_sep": "*",         #opt, defaults "."
                "max_consumers": 32,    #opt, defaults 32
                "topics": [
                    "topic1",
                    "topic2"
                ],
                "servers": [
                    "server.target.com:1234"
                ],
                "max_backlog": 200,     #opt, defaults 200
                "min_backlog": 100      #opt, defaults 50
            }

        Args:
             configs (list[dict]): A list of configs to validate
        Returns:
            bool: True iff all configs are formatted correctly

        """
        required_keys = {"name": str, "source": str, "topics": list, "servers": list, "key_aliases": dict}
        opt_keys = {"key_sep": str, "max_consumers": int, "min_backlog": int, "max_backlog": int}
        for config in configs:
            for key, key_type in required_keys.items():
                if not config.get(key) and not isinstance(config.get(key), key_type):
                    self.ioc.getLogger().error("Config: {0} has an invalid key: {1}".format(config.get('name'), key), notify=True)
                    return False
                if key_type in (list, dict) and len(config.get(key)) == 0:
                    self.ioc.getLogger().error("Config: {0} has an invalid key: {1}".format(config.get('name'), key), notify=True)
                    return False

            for key, key_type in opt_keys.items():
                if config.get(key) and not isinstance(config.get(key), key_type):
                    self.ioc.getLogger().error("Config: {0} has an invalid key: {1}".format(config.get('name'), key), notify=True)
                    return False
                if config.get(key) and key_type in (list, dict) and len(config.get(key)) == 0:
                    self.ioc.getLogger().error("Config: {0} has an invalid key: {1}".format(config.get('name'), key), notify=True)
                    return False

            if config.get("source") != "kafka":
                self.ioc.getLogger().error("Config: {0} is not a kafka config, but it has been loaded by KafkaSource".format(config.get('name')), notify=True)
                return False

            aliases = list(config.get("key_aliases").values())
            if len(aliases) != len(set(aliases)): # if there is a duplicate alias, it is an invalid config
                self.ioc.getLogger().error("Config: {0} has duplicate key_aliases".format(config.get('name')), notify=True)
                return False

        return True



from unittest import TestCase
from tgt_grease.core import Configuration, Logging
import os


class TestLogging(TestCase):
    def test_logging_creation_default(self):
        log = Logging()
        self.assertTrue(isinstance(log, Logging))

    def test_logging_creation_with_conf(self):
        conf = Configuration()
        log = Logging(conf)
        self.assertTrue(isinstance(log, Logging))

    def test_config_getter(self):
        log = Logging()
        self.assertTrue(isinstance(log.getConfig(), Configuration))

    def test_debug(self):
        log = Logging()
        initial = self._log_line_count()
        self.assertTrue(log.debug("Test Debug Message", {'key': 'value'}))
        after = self._log_line_count()
        self.assertTrue(initial + 1 == after)

    def test_info(self):
        log = Logging()
        initial = self._log_line_count()
        self.assertTrue(log.info("Test Info Message", {'key': 'value'}))
        after = self._log_line_count()
        self.assertTrue(initial + 1 == after)

    def test_warning(self):
        log = Logging()
        initial = self._log_line_count()
        self.assertTrue(log.warning("Test Warning Message", {'key': 'value'}))
        after = self._log_line_count()
        self.assertTrue(initial + 1 == after)

    def test_error(self):
        log = Logging()
        initial = self._log_line_count()
        self.assertTrue(log.error("Test Error Message", {'key': 'value'}))
        after = self._log_line_count()
        self.assertTrue(initial + 1 == after)

    def test_critical(self):
        log = Logging()
        initial = self._log_line_count()
        self.assertTrue(log.critical("Test Critical Message", {'key': 'value'}))
        after = self._log_line_count()
        self.assertTrue(initial + 1 == after)

    def _log_line_count(self):
        conf = Configuration()
        if os.path.isfile(conf.get('Logging', 'file')):
            fil = open(conf.get('Logging', 'file'), 'r')
            filContent = fil.read().splitlines()
            fil.close()
            del fil
            ctn = 0
            for line in filContent:
                ctn += 1
            del filContent
            del conf
            return ctn
        else:
            del conf
            return 0



from unittest import TestCase
from tgt_grease.core import GreaseContainer
from tgt_grease.enterprise.Model import PrototypeConfig
from tgt_grease.enterprise.Prototype import scan, detect, schedule
from tgt_grease.router.Commands.Daemon import DaemonProcess
from bson.objectid import ObjectId
import json
import os
import time


class TestFullStack(TestCase):

    def test_mock(self):
        #############################################
        #            SETUP UP TIME
        #############################################
        ioc = GreaseContainer()
        pConf = PrototypeConfig(ioc)
        ioc.ensureRegistration()
        ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': ['scan', 'detect', 'schedule']
                }
            }
        )
        ioc.getCollection('Configuration').insert_one(
            {
                'active': True,
                'type': 'prototype_config',
                "name": "full_stack_test",
                "job": "help",
                "exe_env": "general",
                "source": "url_source",
                "logic": {
                    "Regex": [
                        {
                            "field": "url",
                            "pattern": ".*",
                            'variable': True,
                            'variable_name': 'url'
                        }
                    ],
                    'Range': [
                        {
                            'field': 'status_code',
                            'min': 199,
                            'max': 201
                        }
                    ]
                },
                'constants': {
                    'test': 'ver'
                }
            }
        )
        pConf.load(reloadConf=True)
        TestFile = ioc.getConfig().get('Sourcing', 'dir') + 'full_stack.mock.url.json'
        with open(TestFile, 'w') as fil:
            fil.write(json.dumps({
                'url': 'https://google.com',
                'status_code': 200,
                'headers': str({'test': 'var'}),
                'body': 'test data'
            }, indent=4))
        #############################################
        #            EXECUTE SCANNING
        #############################################
        Scanner = scan()
        Scanner.ioc.getLogger().getConfig().set('verbose', True, 'Logging')
        Scanner.ioc.getLogger().getConfig().set('trace', True, 'Logging')
        Scanner.ioc.getLogger().getConfig().set('mock', True, 'Sourcing')
        Scanner.ioc.getLogger().getConfig().set('config', 'full_stack_test', 'Sourcing')
        self.assertTrue(Scanner.execute({'loop': 1}))
        #############################################
        #            ASSERT SCANNING
        #############################################
        self.assertTrue(ioc.getCollection('SourceData').find_one({
            'grease_data.sourcing.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.start': None,
            'grease_data.detection.end': None
        }))
        #############################################
        #            EXECUTE DETECTION
        #############################################
        Detect = detect()
        Detect.ioc.getLogger().getConfig().set('verbose', True, 'Logging')
        Detect.ioc.getLogger().getConfig().set('trace', True, 'Logging')
        Detect.ioc.getLogger().getConfig().set('mock', True, 'Sourcing')
        Detect.ioc.getLogger().getConfig().set('config', 'full_stack_test', 'Sourcing')
        self.assertTrue(Detect.execute({'loop': 1}))
        #############################################
        #            ASSERT DETECTION
        #############################################
        self.assertTrue(ioc.getCollection('SourceData').find_one({
            'grease_data.sourcing.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.detection.url': ['https://google.com', ''],
            'grease_data.detection.detection.constants.test': 'ver',
            'grease_data.scheduling.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.scheduling.start': None,
            'grease_data.scheduling.end': None
        }))
        #############################################
        #            EXECUTE SCHEDULING
        #############################################
        Scheduling = schedule()
        Scheduling.ioc.getLogger().getConfig().set('verbose', True, 'Logging')
        Scheduling.ioc.getLogger().getConfig().set('trace', True, 'Logging')
        Scheduling.ioc.getLogger().getConfig().set('mock', True, 'Sourcing')
        Scheduling.ioc.getLogger().getConfig().set('config', 'full_stack_test', 'Sourcing')
        self.assertTrue(Scheduling.execute({'loop': 1}))
        #############################################
        #            ASSERT SCHEDULING
        #############################################
        self.assertTrue(ioc.getCollection('SourceData').find_one({
            'grease_data.sourcing.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.detection.url': ['https://google.com', ''],
            'grease_data.detection.detection.constants.test': 'ver',
            'grease_data.scheduling.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.execution.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.execution.start': None,
            'grease_data.execution.end': None
        }))
        #############################################
        #            EXECUTE JOBS
        #############################################
        ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': []
                }
            }
        )
        Daemon = DaemonProcess(ioc)
        Daemon.ioc.getLogger().getConfig().set('verbose', True, 'Logging')
        Daemon.ioc.getLogger().getConfig().set('trace', True, 'Logging')
        Daemon.ioc.getLogger().getConfig().set('mock', True, 'Sourcing')
        Daemon.ioc.getLogger().getConfig().set('config', 'full_stack_test', 'Sourcing')
        self.assertTrue(Daemon.server())
        self.assertTrue(Daemon.drain_jobs(ioc.getCollection('SourceData')))
        #############################################
        #            ASSERT JOB EXECUTION
        #############################################
        # sleep a few for seconds to let help complete
        time.sleep(5)
        self.assertTrue(ioc.getCollection('SourceData').find_one({
            'grease_data.sourcing.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.detection.url': ['https://google.com', ''],
            'grease_data.detection.detection.constants.test': 'ver',
            'grease_data.scheduling.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.execution.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.execution.commandSuccess': True,
            'grease_data.execution.executionSuccess': True
        }))
        #############################################
        #            CLEAN UP TIME
        #############################################
        ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': []
                }
            }
        )
        ioc.getCollection('Configuration').drop()
        ioc.getCollection('SourceData').drop()
        pConf.load(reloadConf=True)
        os.remove(TestFile)

    def test_real(self):
        #############################################
        #            SETUP UP TIME
        #############################################
        ioc = GreaseContainer()
        pConf = PrototypeConfig(ioc)
        ioc.ensureRegistration()
        ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': ['scan', 'detect', 'schedule']
                }
            }
        )
        ioc.getCollection('Configuration').insert_one(
            {
                'active': True,
                'type': 'prototype_config',
                "name": "full_stack_test",
                "job": "help",
                "exe_env": "general",
                "source": "url_source",
                "url": ['http://google.com'],
                "logic": {
                    "Regex": [
                        {
                            "field": "url",
                            "pattern": ".*",
                            'variable': True,
                            'variable_name': 'url'
                        }
                    ],
                    'Range': [
                        {
                            'field': 'status_code',
                            'min': 199,
                            'max': 201
                        }
                    ]
                },
                'constants': {
                    'test': 'ver'
                }
            }
        )
        pConf.load(reloadConf=True)
        #############################################
        #            EXECUTE SCANNING
        #############################################
        Scanner = scan()
        Scanner.ioc.getLogger().getConfig().set('verbose', True, 'Logging')
        Scanner.ioc.getLogger().getConfig().set('trace', True, 'Logging')
        Scanner.ioc.getLogger().getConfig().set('config', 'full_stack_test', 'Sourcing')
        self.assertTrue(Scanner.execute({'loop': 1}))
        #############################################
        #            ASSERT SCANNING
        #############################################
        self.assertTrue(ioc.getCollection('SourceData').find_one({
            'grease_data.sourcing.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.start': None,
            'grease_data.detection.end': None
        }))
        #############################################
        #            EXECUTE DETECTION
        #############################################
        Detect = detect()
        Detect.ioc.getLogger().getConfig().set('verbose', True, 'Logging')
        Detect.ioc.getLogger().getConfig().set('trace', True, 'Logging')
        Detect.ioc.getLogger().getConfig().set('config', 'full_stack_test', 'Sourcing')
        self.assertTrue(Detect.execute({'loop': 1}))
        #############################################
        #            ASSERT DETECTION
        #############################################
        self.assertTrue(ioc.getCollection('SourceData').find_one({
            'grease_data.sourcing.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.scheduling.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.scheduling.start': None,
            'grease_data.scheduling.end': None
        }))
        #############################################
        #            EXECUTE SCHEDULING
        #############################################
        Scheduling = schedule()
        Scheduling.ioc.getLogger().getConfig().set('verbose', True, 'Logging')
        Scheduling.ioc.getLogger().getConfig().set('trace', True, 'Logging')
        Scheduling.ioc.getLogger().getConfig().set('config', 'full_stack_test', 'Sourcing')
        self.assertTrue(Scheduling.execute({'loop': 1}))
        #############################################
        #            ASSERT SCHEDULING
        #############################################
        self.assertTrue(ioc.getCollection('SourceData').find_one({
            'grease_data.sourcing.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.scheduling.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.execution.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.execution.start': None,
            'grease_data.execution.end': None
        }))
        #############################################
        #            EXECUTE JOBS
        #############################################
        ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': []
                }
            }
        )
        Daemon = DaemonProcess(ioc)
        Daemon.ioc.getLogger().getConfig().set('verbose', True, 'Logging')
        Daemon.ioc.getLogger().getConfig().set('trace', True, 'Logging')
        Daemon.ioc.getLogger().getConfig().set('config', 'full_stack_test', 'Sourcing')
        self.assertTrue(Daemon.server())
        self.assertTrue(Daemon.drain_jobs(ioc.getCollection('SourceData')))
        #############################################
        #            ASSERT JOB EXECUTION
        #############################################
        # sleep a few for seconds to let help complete
        time.sleep(5)
        self.assertTrue(ioc.getCollection('SourceData').find_one({
            'grease_data.sourcing.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.detection.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.scheduling.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.execution.server': ObjectId(ioc.getConfig().NodeIdentity),
            'grease_data.execution.commandSuccess': True,
            'grease_data.execution.executionSuccess': True
        }))
        #############################################
        #            CLEAN UP TIME
        #############################################
        ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': []
                }
            }
        )
        ioc.getCollection('Configuration').drop()
        ioc.getCollection('SourceData').drop()
        ioc.getCollection('DeDup_Sourcing').drop()
        pConf.load(reloadConf=True)



from tgt_grease.core import Configuration
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL
from urllib3.exceptions import HTTPError
import requests
import json



class Notifications(object):
    """Notification Router for third party resources

    This is the class to handle all notifications to third party resources

    Attributes:
        _conf (Configuration): Configuration Object
        hipchat_url (str): This is the hipchat API url
        hipchat_token (str): set this to override the config for the auth token
        hipchat_room (str): set this to override the config for the room

    """

    _conf = None
    # HipChat Configuration
    hipchat_url = "https://api.hipchat.com/v2/room/"
    hipchat_token = None
    hipchat_room = None

    def __init__(self, Config=None):
        if Config and isinstance(Config, Configuration):
            self._conf = Config
        else:
            self._conf = Configuration()

    def SendMessage(self, message, level=DEBUG, channel=None):
        """Send Message to configured channels

        This method is the main point of contact with Notifications in GREASE. This will handle routing to all
        configured channels. Use `level` to define what level the message is. This can impact whether a message is
        sent as well as if the message sent will have special attributes (EX: red text). Use `channel` to route
        around sending to multiple channels if the message traditionally would go to multiple, instead going only
        to the selected one.

        Note:
            if you use the channel argument and the channel is not found you will receive False back

        Args:
            message (str): Message to send
            level (int): Level of message to be sent
            channel (str): Specific channel to notify

        Returns:
            bool: Success of sending

        """
        if channel \
                and 'enabled' in self._conf.get('Notifications', channel, {}) \
                and self._conf.get('Notifications', channel, {}).get('enabled'):
            return bool(self._route_notification(channel, message, level))
        else:
            # Capture object for notification channel statuses
            NotificationStatus = []
            NotificationChannels = self._conf.get('Notifications', default={})  # type: dict
            # Loop through those channels
            for Notifier, Config in NotificationChannels.items():
                # ensure channel is enabled
                if 'enabled' in Config and Config.get('enabled'):
                    # loop through the channels
                    NotificationStatus.append(bool(self._route_notification(Notifier, message, level)))
            # make the list unique
            NotificationStatus = list(set(NotificationStatus))
            if len(NotificationStatus) > 1:
                # we got at least one true and at least one false
                return False
            elif len(NotificationStatus) is 1:
                # return what the categorical state was
                return bool(NotificationStatus[0])
            else:
                # nothing was configured to run return true
                return True

    def _route_notification(self, channel, message, level):
        """Handle actual calling of notification channels

        Args:
            channel (str): Channel to notify
            message (str): Message to send
            level (int): Level to send at

        Returns:
            bool: Method success status

        """
        if channel == "HipChat":
            return self.send_hipchat_message(message, level)
        elif channel == "Slack":
            return self.send_slack_message(message)
        else:
            return False

    def send_hipchat_message(self, message, level, color=None):
        """Send a hipchat message

        Args:
            message (str): Message to send to hipchat
            level (int): message level
            color (str): color of message

        Returns:
            bool: API response status

        """
        if not color:
            if level is DEBUG:
                color = 'grey'
            elif level is INFO:
                color = 'purple'
            elif level is WARNING:
                color = 'yellow'
            elif level is ERROR:
                color = 'red'
            elif level is CRITICAL:
                color = 'red'
            else:
                color = 'grey'
        if self.hipchat_room and self.hipchat_token:
            url = "{0}{1}/notification?auth_token={2}".format(
                self.hipchat_url,
                self.hipchat_room,
                self.hipchat_token
            )
        else:
            url = "{0}{1}/notification?auth_token={2}".format(
                self.hipchat_url,
                self._conf.get('Notifications', 'HipChat').get('room'),
                self._conf.get('Notifications', 'HipChat').get('token')
            )
        try:
            response = requests.post(
                url=url,
                data={
                    'message': message,
                    'color': color
                },
                verify=False
                )
            if response.status_code == 204:
                return True
            else:
                return False
        except HTTPError:
            return False

    def send_slack_message(self, message):
        """Send a slack message to slack channel using webhook url in the configuration

        Args:
            message (str): Message to send to Slack

        Returns:
            bool: API response status

        """

        webhook_url = self._conf.get('Notifications', 'Slack').get('webhookURL')
        slack_data = {'text': message}

        try:
            response = requests.post(
                webhook_url, data=json.dumps(slack_data),
                headers={'Content-Type': 'application/json'}
            )
            return True
        except HTTPError:
            return False




from tgt_grease.enterprise.Model import BaseSourceClass
from tgt_grease.core import Configuration, GreaseContainer
import psycopg2
from psycopg2.extras import RealDictCursor
import datetime
import fnmatch
import json
import os


class SQLSource(BaseSourceClass):
    """Source data from a SQL Database

    This Source is designed to query a SQL Server for data. A generic configuration looks like this for a
    sql_source::

        {
            'name': 'example_source', # <-- A name
            'job': 'example_job', # <-- Any job you want to run
            'exe_env': 'general', # <-- Selected execution environment; Can be anything!
            'source': 'sql_source', # <-- This source
            'type': 'postgresql', # <-- SQL Server Type (Only supports PostgreSQL Currently)
            'dsn': 'SQL_SERVER_CONNECTION', # <-- String representing the Environment variable used to connect with
            'query': 'select count(*) as order_total from orders where oDate::DATE = current_data', # <-- SQL Query to execute on server
            'hour': 16, # <-- **OPTIONAL** 24hr time hour to poll SQL
            'minute': 30, # <-- **OPTIONAL** Minute to poll SQL
            'logic': {} # <-- Whatever logic your heart desires
        }

    Note:
        This configuration is an example
    Note:
        Currently We only support PostreSQL Server
    Note:
        without `minute` parameter the engine will poll for the entire hour
    Note:
        **Hour and minute parameters are in UTC time**
    Note:
        To only poll once an hour only set the **minute** field

    """

    def parse_source(self, configuration):
        """This will Query the SQL Server to find data

        Args:
            configuration (dict): Configuration of Source. See Class Documentation above for more info

        Returns:
            bool: If True data will be scheduled for ingestion after deduplication. If False the engine will bail out

        """
        ioc = GreaseContainer()
        if configuration.get('hour'):
            if datetime.datetime.utcnow().hour != int(configuration.get('hour')):
                # it is not the correct hour
                return True
        if configuration.get('minute'):
            if datetime.datetime.utcnow().minute != int(configuration.get('minute')):
                # it is not the correct hour
                return True
        if configuration.get('type') != 'postgresql':
            ioc.getLogger().error("Unsupported SQL Server Type; Currently Only supporting PostgreSQL", notify=False)
            return False
        else:
            # Attempt to get the DSN for the connection
            if os.environ.get(configuration.get('dsn')) and configuration.get('query'):
                # ensure the DSN is setup and the query is present
                try:
                    DSN = os.environ.get(configuration.get('dsn'))
                    with psycopg2.connect(DSN) as conn:
                        with conn.cursor(cursor_factory=RealDictCursor) as cursor:
                            cursor.execute(configuration.get('query'))
                            data = cursor.fetchall()
                            for row in data:
                                self._data.append(row)
                            del ioc
                    return True
                except Exception as e:
                    # Naked except to prevent issues around connections
                    ioc.getLogger().error("Error processing configuration; Error [{0}]".format(e.message), notify=False)
                    del ioc
                    return False
            else:
                # could not get the DSN
                ioc.getLogger().error("Failed to locate the DSN variable", notify=False)
                del ioc
                return False

    def mock_data(self, configuration):
        """Data from this source is mocked utilizing the GREASE Filesystem

        Mock data for this source can be place in `<GREASE_DIR>/etc/*.mock.sql.json`. This source will pick up all these
        files and load them into the returning object. The data in these files should reflect what you expect to return
        from SQL::

            {
                'column expected': 'value expected'
                ...
            }

        Args:
            configuration (dict): Configuration Data for source

        Note:
            Argument **configuration** is not honored here
        Note:
            A mock file should represent a single row

        Returns:
            list[dict]: Mocked Data

        """
        intermediate = list()
        matches = []
        conf = Configuration()
        for root, dirnames, filenames in os.walk(conf.greaseDir + 'etc'):
            for filename in fnmatch.filter(filenames, '*.mock.sql.json'):
                matches.append(os.path.join(root, filename))
        for doc in matches:
            with open(doc) as current_file:
                content = current_file.read().replace('\r\n', '')
            try:
                intermediate.append(json.loads(content))
            except ValueError:
                continue
        return intermediate



from unittest import TestCase
from tgt_grease.enterprise.Model import Scheduler, PrototypeConfig, Detect, Deduplication
import datetime
from bson.objectid import ObjectId
import pymongo
import platform


class TestScheduler(TestCase):

    def test_schedule_good(self):
        d = Detect()
        p = PrototypeConfig(d.ioc)
        s = Scheduler(d.ioc)
        source = {
            'key': 'var',
            'ver': 'key',
            'greg': 'old',
            'created': '2017-11-24'
        }
        configuration = {
            'name': 'demoConfig',
            'job': 'otherThing',
            'exe_env': 'general',
            'source': 'Google',
            'logic': {
                'DateRange': [
                    {
                        'field': 'created',
                        'format': '%Y-%m-%d',
                        'min': '2017-11-23',
                        'max': '2017-11-25'
                    }
                ],
                'Regex': [
                    {
                        'field': 'key',
                        'pattern': '.*',
                        'variable': True,
                        'variable_name': 'field'
                    },
                    {
                        'field': 'ver',
                        'pattern': '.*'
                    }
                ],
                'Exists': [
                    {
                        'field': 'greg',
                        'variable': True,
                        'variable_name': 'greg'
                    }
                ]
            }
        }
        p.load(True, [configuration])
        sourceId = d.ioc.getCollection('SourceData').insert_one({
                    'grease_data': {
                        'sourcing': {
                            'server': ObjectId(d.ioc.getConfig().NodeIdentity)
                        },
                        'detection': {
                            'server': ObjectId(d.ioc.getConfig().NodeIdentity),
                            'detectionStart': None,
                            'detectionEnd': None,
                            'detection': {}
                        },
                        'scheduling': {
                            'schedulingServer': None,
                            'schedulingStart': None,
                            'schedulingEnd': None
                        },
                        'execution': {
                            'server': None,
                            'assignmentTime': None,
                            'executionStart': None,
                            'executionEnd': None,
                            'context': {},
                            'executionSuccess': False,
                            'commandSuccess': False,
                            'failures': 0,
                            'retryTime': datetime.datetime.utcnow()
                        }
                    },
                    'source': str('test').encode('utf-8'),
                    'configuration': configuration.get('name'),
                    'data': source,
                    'createTime': datetime.datetime.utcnow(),
                    'expiry': Deduplication.generate_max_expiry_time(1)
        }).inserted_id
        d.ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(d.ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': ["scan", "detect", "schedule"]
                }
            }
        )
        self.assertTrue(d.detectSource())
        self.assertFalse(d.getScheduledSource())
        self.assertTrue(s.getDetectedSource())
        self.assertTrue(s.schedule(s.ioc.getCollection('SourceData').find_one({'_id': ObjectId(sourceId)})))
        self.assertTrue(d.ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(d.ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': []
                }
            }
        ))
        d.ioc.getCollection('SourceData').delete_one({'_id': ObjectId(sourceId)})

    def test_scheduling(self):
        d = Detect()
        p = PrototypeConfig(d.ioc)
        s = Scheduler(d.ioc)
        source = {
            'key': 'var',
            'ver': 'key',
            'greg': 'old',
            'created': '2017-11-24'
        }
        configuration = {
            'name': 'demoConfig',
            'job': 'otherThing',
            'exe_env': 'general',
            'source': 'Google',
            'logic': {
                'DateRange': [
                    {
                        'field': 'created',
                        'format': '%Y-%m-%d',
                        'min': '2017-11-23',
                        'max': '2017-11-25'
                    }
                ],
                'Regex': [
                    {
                        'field': 'key',
                        'pattern': '.*',
                        'variable': True,
                        'variable_name': 'field'
                    },
                    {
                        'field': 'ver',
                        'pattern': '.*'
                    }
                ],
                'Exists': [
                    {
                        'field': 'greg',
                        'variable': True,
                        'variable_name': 'greg'
                    }
                ]
            }
        }
        p.load(True, [configuration])
        sourceId = d.ioc.getCollection('SourceData').insert_one({
                    'grease_data': {
                        'sourcing': {
                            'server': ObjectId(d.ioc.getConfig().NodeIdentity)
                        },
                        'detection': {
                            'server': ObjectId(d.ioc.getConfig().NodeIdentity),
                            'detectionStart': None,
                            'detectionEnd': None,
                            'detection': {}
                        },
                        'scheduling': {
                            'schedulingServer': None,
                            'schedulingStart': None,
                            'schedulingEnd': None
                        },
                        'execution': {
                            'server': None,
                            'assignmentTime': None,
                            'executionStart': None,
                            'executionEnd': None,
                            'context': {},
                            'executionSuccess': False,
                            'commandSuccess': False,
                            'failures': 0,
                            'retryTime': datetime.datetime.utcnow()
                        }
                    },
                    'source': str('test').encode('utf-8'),
                    'configuration': configuration.get('name'),
                    'data': source,
                    'createTime': datetime.datetime.utcnow(),
                    'expiry': Deduplication.generate_max_expiry_time(1)
        }).inserted_id
        d.ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(d.ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': ["scan", "detect", "schedule"]
                }
            }
        )
        self.assertTrue(d.detectSource())
        self.assertFalse(d.getScheduledSource())
        self.assertTrue(s.getDetectedSource())
        self.assertTrue(s.scheduleExecution())
        self.assertTrue(d.ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(d.ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': []
                }
            }
        ))
        d.ioc.getCollection('SourceData').delete_one({'_id': ObjectId(sourceId)})

    def test_scheduling_fail(self):
        d = Detect()
        p = PrototypeConfig(d.ioc)
        s = Scheduler(d.ioc)
        source = {
            'key': 'var',
            'ver': 'key',
            'greg': 'old',
            'created': '2017-11-24'
        }
        configuration = {
            'name': 'demoConfig',
            'job': 'otherThing',
            'exe_env': 'minix',
            'source': 'Google',
            'logic': {
                'DateRange': [
                    {
                        'field': 'created',
                        'format': '%Y-%m-%d',
                        'min': '2017-11-23',
                        'max': '2017-11-25'
                    }
                ],
                'Regex': [
                    {
                        'field': 'key',
                        'pattern': '.*',
                        'variable': True,
                        'variable_name': 'field'
                    },
                    {
                        'field': 'ver',
                        'pattern': '.*'
                    }
                ],
                'Exists': [
                    {
                        'field': 'greg',
                        'variable': True,
                        'variable_name': 'greg'
                    }
                ]
            }
        }
        p.load(True, [configuration])
        sourceId = d.ioc.getCollection('SourceData').insert_one({
                    'grease_data': {
                        'sourcing': {
                            'server': ObjectId(d.ioc.getConfig().NodeIdentity)
                        },
                        'detection': {
                            'server': ObjectId(d.ioc.getConfig().NodeIdentity),
                            'detectionStart': None,
                            'detectionEnd': None,
                            'detection': {}
                        },
                        'scheduling': {
                            'schedulingServer': None,
                            'schedulingStart': None,
                            'schedulingEnd': None
                        },
                        'execution': {
                            'server': None,
                            'assignmentTime': None,
                            'executionStart': None,
                            'executionEnd': None,
                            'context': {},
                            'executionSuccess': False,
                            'commandSuccess': False,
                            'failures': 0,
                            'retryTime': datetime.datetime.utcnow()
                        }
                    },
                    'source': str('test').encode('utf-8'),
                    'configuration': configuration.get('name'),
                    'data': source,
                    'createTime': datetime.datetime.utcnow(),
                    'expiry': Deduplication.generate_max_expiry_time(1)
        }).inserted_id
        d.ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(d.ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': ["scan", "detect", "schedule"]
                }
            }
        )
        self.assertTrue(d.detectSource())
        self.assertFalse(d.getScheduledSource())
        self.assertTrue(s.getDetectedSource())
        self.assertFalse(s.scheduleExecution())
        self.assertTrue(d.ioc.getCollection('JobServer').update_one(
            {'_id': ObjectId(d.ioc.getConfig().NodeIdentity)},
            {
                '$set': {
                    'prototypes': []
                }
            }
        ))
        d.ioc.getCollection('SourceData').delete_one({'_id': ObjectId(sourceId)})



from tgt_grease.enterprise.Model import BaseSourceClass
from tgt_grease.core import Configuration
import json
import fnmatch
import os
import requests
import datetime


class URLParser(BaseSourceClass):
    """Monitor URL's as a source of information

    This source is designed to provide source data on the URL's configured for a GREASE sourcing cluster. A generic
    configuration looks like this for a url_source::

        {
            'name': 'example_source', # <-- A name
            'job': 'example_job', # <-- Any job you want to run
            'exe_env': 'general', # <-- Selected execution environment; Can be anything!
            'source': 'url_source', # <-- This source
            'url': ['google.com', 'http://bing.com', '8.8.8.8'], # <-- List of URL's to parse
            'hour': 16, # <-- **OPTIONAL** 24hr time hour to poll URLs
            'minute': 30, # <-- **OPTIONAL** Minute to poll URLs
            'logic': {} # <-- Whatever logic your heart desires
        }

    Note:
        This configuration is an example
    Note:
        If a URL in the `url` parameter is not prefixed with `http://` then the class will do so for you
    Note:
        without `minute` parameter the engine will poll for the entire hour
    Note:
        **Hour and minute parameters are in UTC time**
    Note:
        To only poll once an hour only set the **minute** field

    """

    def parse_source(self, configuration):
        """This will make a GET request to all URL's in the list provided by your configuration

        Args:
            configuration (dict): Configuration of Source. See Class Documentation above for more info

        Returns:
            bool: If True data will be scheduled for ingestion after deduplication. If False the engine will bail out

        """
        scannable = len(configuration.get('url', []))
        if scannable is 0:
            return False
        scanned = 0
        if configuration.get('hour'):
            if datetime.datetime.utcnow().hour != int(configuration.get('hour')):
                # it is not the correct hour
                return True
        if configuration.get('minute'):
            if datetime.datetime.utcnow().minute != int(configuration.get('minute')):
                # it is not the correct hour
                return True
        for URL in configuration.get('url', []):  # type: str
            if not URL.startswith("http://"):
                URL = "http://" + URL
            try:
                response = requests.get(URL)
                self._data.append({
                    'url': response.url.encode('utf-8', 'ignore'),
                    'status_code': int(response.status_code),
                    'headers': str(response.headers),
                    'body': response.text.encode('utf-8', 'ignore')
                })
                scanned += 1
            except requests.HTTPError:
                continue
        if scanned < (scannable / 2):
            return False
        else:
            return True

    def mock_data(self, configuration):
        """Data from this source is mocked utilizing the GREASE Filesystem

        Mock data for this source can be place in `<GREASE_DIR>/etc/*.mock.url.json`. This source will pick up all these
        files and load them into the returning object. They will need to follow this schema::

            {
                'url': String, # <-- URL that would have been loaded
                'status_code': Int, # <-- HTTP Status code
                'headers': String, # <-- HTTP headers as a string
                'body': String # <-- HTTP response body
            }

        Args:
            configuration (dict): Configuration Data for source

        Note:
            Argument **configuration** is not honored here

        Returns:
            list[dict]: Mocked Data

        """
        intermediate = list()
        matches = []
        conf = Configuration()
        for root, dirnames, filenames in os.walk(conf.greaseDir + 'etc'):
            for filename in fnmatch.filter(filenames, '*.mock.url.json'):
                matches.append(os.path.join(root, filename))
        for doc in matches:
            with open(doc) as current_file:
                content = current_file.read().replace('\r\n', '')
            try:
                intermediate.append(json.loads(content))
            except ValueError:
                continue
        return intermediate



from tgt_grease.core import GreaseContainer
from tgt_grease.core import ImportTool
from .Configuration import PrototypeConfig
from .CentralScheduling import Scheduling
from .BaseDetector import Detector
from bson.objectid import ObjectId
import pymongo
import datetime


class Detect(object):
    """Detection class for GREASE detect

    This is the model to actually utilize the detectors to parse the sources from scan

    Attributes:
        ioc (GreaseContainer): IOC for scanning
        impTool (ImportTool): Import Utility Instance
        conf (PrototypeConfig): Prototype configuration tool
        scheduler (Scheduling): Prototype Scheduling Service Instance

    """

    def __init__(self, ioc=None):
        if ioc and isinstance(ioc, GreaseContainer):
            self.ioc = ioc
        else:
            self.ioc = GreaseContainer()
        self.impTool = ImportTool(self.ioc.getLogger())
        self.ioc.ensureRegistration()
        self.conf = PrototypeConfig(self.ioc)
        self.scheduler = Scheduling(self.ioc)

    def detectSource(self):
        """This will perform detection the oldest source from SourceData

        Returns:
            bool: If detection process was successful

        """
        sourceData = self.getScheduledSource()
        if sourceData:
            if isinstance(sourceData.get('configuration'), bytes):
                conf = sourceData.get('configuration').decode()
            else:
                conf = sourceData.get('configuration')
            configurationData = self.conf.get_config(conf)
            if configurationData:
                self.ioc.getCollection('SourceData').update_one(
                    {'_id': ObjectId(sourceData.get('_id'))},
                    {
                        '$set': {
                            'grease_data.detection.start': datetime.datetime.utcnow()
                        }
                    }
                )
                result, resultData = self.detection(sourceData.get('data'), configurationData)
                if result:
                    # Put constants in detection results
                    resultData['constants'] = self.conf.get_config(configurationData.get('name')).get('constants', {})
                    # Update detection
                    self.ioc.getCollection('SourceData').update_one(
                        {'_id': ObjectId(sourceData.get('_id'))},
                        {
                            '$set': {
                                'grease_data.detection.end': datetime.datetime.utcnow(),
                                'grease_data.detection.detection': resultData
                            }
                        }
                    )
                    # attempt scheduling
                    return self.scheduler.scheduleScheduling(sourceData.get('_id'))
                else:
                    self.ioc.getCollection('SourceData').update_one(
                        {'_id': ObjectId(sourceData.get('_id'))},
                        {
                            '$set': {
                                'grease_data.detection.end': datetime.datetime.utcnow(),
                                'grease_data.detection.detection': {}
                            }
                        }
                    )
                    self.ioc.getLogger().trace("Detection yielded no detection data", trace=True)
                    return True
            else:
                self.ioc.getLogger().error(
                    "Failed to load Prototype Config [{0}]".format(sourceData.get('configuration')),
                    notify=False
                )
                return False
        else:
            self.ioc.getLogger().trace("No sources awaiting detection currently", trace=True)
            return True

    def getScheduledSource(self):
        """Queries for oldest source that has been assigned for detection

        Returns:
            dict: source awaiting detection

        """
        return self.ioc.getCollection('SourceData').find_one(
            {
                'grease_data.detection.server': ObjectId(self.ioc.getConfig().NodeIdentity),
                'grease_data.detection.start': None,
                'grease_data.detection.end': None,
            },
            sort=[('grease_data.createTime', pymongo.DESCENDING)]
        )

    def detection(self, source, configuration):
        """Performs detection on a source with the provided configuration

        Args:
            source (dict): Key->Value pairs from sourcing to detect upon
            configuration (dict): Prototype configuration provided from sourcing

        Returns:
            tuple: Detection Results; first boolean for success, second dict of variables for context

        """
        # Ensure types
        final = {}
        finalBool = False
        if not isinstance(source, dict):
            self.ioc.getLogger().warning("Detection got non-dict source data", notify=False)
            finalBool = False
            return finalBool, final
        if not isinstance(configuration, dict):
            self.ioc.getLogger().warning("Detection got non-dict configuration", notify=False)
            finalBool = False
            return finalBool, final
        # Now loop through logical blocks
        for detector, logicBlock in configuration.get('logic', {}).items():
            if not isinstance(logicBlock, list):
                self.ioc.getLogger().warning("Logical Block was not list", trace=True, notify=False)
            detect = self.impTool.load(detector)
            if isinstance(detect, Detector):
                result, resultData = detect.processObject(source, logicBlock)
                if not result:
                    self.ioc.getLogger().trace("Detection yielded false for [{0}]".format(detector), trace=True)
                    finalBool = False
                    break
                else:
                    self.ioc.getLogger().trace("Detection yielded true for [{0}]".format(detector), trace=True)
                    for key, val in resultData.items():
                        final[key] = val
                    finalBool = True
                    continue
            else:
                self.ioc.getLogger().warning("invalid detector [{0}]".format(detector), notify=False)
                finalBool = False
        return finalBool, final



from tgt_grease.enterprise.Model import Detector
import datetime


class DateRange(Detector):
    """Date Range Detector for GREASE Detection

    A Typical DateRange configuration looks like this::

        {
            ...
            'logic': {
                'DateRange': [
                    {
                        'field': String, # <-- Field to search for
                        'min': DateTime String, # <-- OPTIONAL IF max is set
                        'max': DateTime String, # <-- OPTIONAL IF min is set
                        'format': '%Y-%m-%d', # <-- Mandatory via strptime behavior
                        'variable': Boolean, # <-- OPTIONAL, if true then create a context variable of result
                        'variable_name: String # <-- REQUIRED IF variable, name of context variable
                    }
                    ...
                ]
                ...
            }
        }

    Note:
        The min field is only required if max is not present. This would ensure the field is after the date provided
    Note:
        The max field is only required if min is not present. This would ensure the field is before the date provided
    Note:
        Change the format to any supported https://docs.python.org/2/library/datetime.html#strftime-and-strptime-behavior
    Note:
        To find an exact date/time set min one below the target and max one above the target

    """

    def processObject(self, source, ruleConfig):
        """Processes an object and returns valid rule data

        Data returned in the second parameter from this method should be in this form::

            {
                '<field>': Object # <-- if specified as a variable then return the key->Value pairs
                ...
            }

        Args:
            source (dict): Source Data
            ruleConfig (list[dict]): Rule Configuration Data

        Return:
            tuple: first element boolean for success; second dict for any fields returned as variables

        """
        finalBool = False
        final = {}
        # type checks
        if not isinstance(source, dict):
            return False, {}
        if not isinstance(ruleConfig, list):
            return False, {}
        for block in ruleConfig:
            if not isinstance(block, dict):
                self.ioc.getLogger().error(
                    "INVALID DATERANGE LOGICAL BLOCK! NOT TYPE LIST [{0}]".format(str(type(block))),
                    notify=False
                )
                return False, {}
            # ensure field is there and date format
            if not block.get('field') in source or 'format' not in block:
                self.ioc.getLogger().error(
                    "malformed rule block; field and/or format not found in source",
                    notify=False
                )
                return False, {}
            if not source.get(block.get('field')):
                self.ioc.getLogger().error(
                    "field equated to False!",
                    notify=False
                )
                return False, {}
            if self.timeCompare(source.get(block.get('field')), block):
                finalBool = True
                if block.get('variable') and block.get('variable_name'):
                    final[str(block.get('variable_name'))] = source.get(block.get('field'))
                else:
                    continue
            else:
                self.ioc.getLogger().trace("Field Failed Range Comparison", verbose=True)
                return False, {}
        return finalBool, final

    def timeCompare(self, field, LogicalBlock):
        """Compares a date to a range

        Args:
            field (str): field to compare
            LogicalBlock (dict): Logical Block

        Returns:
            bool: if the range is successful then true else false

        """
        try:
            source_date = datetime.datetime.strptime(field, LogicalBlock.get('format'))
            # Ensure field is present
            # ensure at least min OR max is present
            if not LogicalBlock.get('min') and not LogicalBlock.get('max'):
                self.ioc.getLogger().trace("[min] and/or [max] not found in config block", verbose=True)
                return False
            if LogicalBlock.get('min') and LogicalBlock.get('max'):
                # Min & Max Defined
                if datetime.datetime.strptime(LogicalBlock.get('min'), LogicalBlock.get('format')) <= source_date <= datetime.datetime.strptime(LogicalBlock.get('max'), LogicalBlock.get('format')):
                    return True
                else:
                    return False
            elif LogicalBlock.get('min') and not LogicalBlock.get('max'):
                # Min Defined
                if source_date >= datetime.datetime.strptime(LogicalBlock.get('min'), LogicalBlock.get('format')):
                    return True
                else:
                    return False
            elif not LogicalBlock.get('min') and LogicalBlock.get('max'):
                # Max Defined
                if source_date <= datetime.datetime.strptime(LogicalBlock.get('max'), LogicalBlock.get('format')):
                    return True
                else:
                    return False
            else:
                self.ioc.getLogger().error("Failed to find either min OR max in LogicalBlock", verbose=True, notify=False)
                return False
        except ValueError:
            # probable datetime format error
            self.ioc.getLogger().error("Value error processing rule!", notify=False)
            return False
        except TypeError:
            # probable datetime format error
            self.ioc.getLogger().error("Type error processing rule!", notify=False)
            return False



from tgt_grease.router import GreaseRouter
from tgt_grease.router.Commands import help
import sys
from unittest import TestCase


class TestRouter(TestCase):
    def test_no_SubCommand(self):
        sys.argv = ['grease']
        cmd = GreaseRouter()
        self.assertEqual(cmd.run(), 1)

    def test_get_cli_args_without_command(self):
        sys.argv = ['grease', '--text=utf-8', '--opt', 'var', '--ver:var']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertIsNone(cmd)
        self.assertDictEqual(context, {'opt': 'var', 'text': 'utf-8', 'ver': 'var', 'grease_other_args': []})

    def test_get_cli_args_with_command_type1(self):
        sys.argv = ['grease', '--text=utf-8', '--opt', 'var', '--ver:var', 'help']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(context, {'opt': 'var', 'text': 'utf-8', 'ver': 'var', 'grease_other_args': []})

    def test_get_cli_args_with_command_type2(self):
        sys.argv = ['grease', 'help', '--text=utf-8', '--opt', 'var', '--ver:var']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(context, {'opt': 'var', 'text': 'utf-8', 'ver': 'var', 'grease_other_args': []})

    def test_get_cli_args_with_command_type3(self):
        sys.argv = ['grease', '--text=utf-8', 'help', '--opt', 'var', '--ver:var']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(context, {'opt': 'var', 'text': 'utf-8', 'ver': 'var', 'grease_other_args': []})

    def test_get_cli_args_with_command_type4(self):
        sys.argv = ['grease', '--text=utf-8', '--opt', 'var', 'help', '--ver:var']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(context, {'opt': 'var', 'text': 'utf-8', 'ver': 'var', 'grease_other_args': []})

    def test_get_cli_args_with_command_type5(self):
        sys.argv = ['grease', 'help', '--text=utf-8', '--opt', 'var', '--ver:var', 'install']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(context, {'opt': 'var', 'text': 'utf-8', 'ver': 'var', 'grease_other_args': ['install']})

    def test_get_cli_args_with_command_type6(self):
        sys.argv = ['grease', 'help', 'install', '--text=utf-8', '--opt', 'var', '--ver:var']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(context, {'opt': 'var', 'text': 'utf-8', 'ver': 'var', 'grease_other_args': ['install']})

    def test_get_cli_args_with_command_type7(self):
        sys.argv = ['grease', 'help', '--text=utf-8', 'install', '--opt', 'var', '--ver:var']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(context, {'opt': 'var', 'text': 'utf-8', 'ver': 'var', 'grease_other_args': ['install']})

    def test_get_cli_args_with_command_type8(self):
        sys.argv = ['grease', 'help', 'install']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(context, {'grease_other_args': ['install']})

    def test_get_cli_args_with_command_type9(self):
        sys.argv = ['grease', 'help', 'install', 'to-do-stuff']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(context, {'grease_other_args': ['install', 'to-do-stuff']})

    def test_get_cli_args_with_command_type10(self):
        sys.argv = ['grease', 'help', 'install', 'to-do-stuff', '--foreground']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(context, {'foreground': True, 'grease_other_args': ['install', 'to-do-stuff']})

    def test_get_cli_args_with_command_type11(self):
        sys.argv = ['grease', 'help', '--test=var', '--foreground', '--test1:var1', '--test2', 'var2', 'install']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertDictEqual(
            {'foreground': True, 'test': 'var', 'test1': 'var1', 'test2': 'var2', 'grease_other_args': ['install']},
            context
        )

    def test_get_cli_args_with_command_type12(self):
        sys.argv = ['grease', 'help', '--test=var', '--foreground', '--test1:var1', '--test2', 'var2']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertTrue(
            context,
            {'foreground': True, 'test': 'var', 'test1': 'var1', 'test2': 'var2', 'grease_other_args': []}
        )

    def test_get_cli_args_with_command_type13(self):
        sys.argv = ['grease', '--foreground', 'help', '--test=var', '--test1:var1', '--test2', 'var2']
        rtr = GreaseRouter()
        cmd, context = rtr.get_arguments()
        self.assertTrue(isinstance(cmd, help))
        self.assertTrue(
            context,
            {'foreground': True, 'test': 'var', 'test1': 'var1', 'test2': 'var2', 'grease_other_args': []}
        )



from .BaseSource import BaseSourceClass
from .BaseDetector import Detector
from .DeDuplication import Deduplication
from .CentralScheduling import Scheduling
from .Scanning import Scan
from .Detection import Detect
from .Scheduler import Scheduler
from .Configuration import PrototypeConfig
from .KafkaSource import KafkaSource



from unittest import TestCase
from tgt_grease.core.Types import Command
from tgt_grease.core import GreaseContainer


class TestCmd(Command):
    def __init__(self):
        super(TestCmd, self).__init__()

    def execute(self, context):
        for key, val in context.items():
            self.setData(key, val)
        return True


class TestCmdFail(Command):
    def __init__(self):
        super(TestCmdFail, self).__init__()

    def execute(self, context):
        for key, val in context.items():
            self.setData(key, val)
        return False


class TestCmdExcept(Command):
    def __init__(self):
        super(TestCmdExcept, self).__init__()

    def execute(self, context):
        for key, val in context.items():
            self.setData(key, val)
        raise AttributeError


class TestCommand(TestCase):
    def test_ioc(self):
        cmd = TestCmd()
        self.assertTrue(isinstance(cmd.ioc, GreaseContainer))

    def test_variable_storage_name(self):
        cmd = TestCmd()
        self.assertEqual(cmd.variable_storage.name, "TestCmd")

    def test_variable_storage(self):
        cmd = TestCmd()
        cmd.variable_storage.insert_one({'test': 'value'})
        self.assertEqual(cmd.variable_storage.find({'test': 'value'}).count(), 1)
        self.assertTrue(cmd.variable_storage.delete_one({'test': 'value'}))

    def test_get_exec_info(self):
        cmd = TestCmd()
        cmd.safe_execute({})
        self.assertTrue(isinstance(cmd.getExecVal(), bool))

    def test_get_ret_val_info(self):
        cmd = TestCmd()
        cmd.safe_execute({})
        self.assertTrue(isinstance(cmd.getRetVal(), bool))

    def test_get_ret_data(self):
        cmd = TestCmd()
        cmd.safe_execute({'key': 'value'})
        self.assertDictEqual(cmd.getData(), {'key': 'value'})

    def test_failed_cmd(self):
        cmd = TestCmdFail()
        cmd.safe_execute({})
        self.assertTrue(cmd.getExecVal())
        self.assertFalse(cmd.getRetVal())

    def test_except_cmd(self):
        cmd = TestCmdExcept()
        cmd.safe_execute({})
        self.assertFalse(cmd.getExecVal())
        self.assertFalse(cmd.getRetVal())



from unittest import TestCase
from tgt_grease.core import GreaseContainer
from tgt_grease.enterprise.Model import Scheduling, Detect, PrototypeConfig, Deduplication
from bson.objectid import ObjectId
import datetime
import platform
import time
import pymongo


class TestScheduling(TestCase):

    def test_empty_source_schedule(self):
        ioc = GreaseContainer()
        sch = Scheduling(ioc)
        jServer = ioc.getCollection('JobServer')
        jID = jServer.insert_one({
                'jobs': 0,
                'os': platform.system().lower(),
                'roles': ["general"],
                'prototypes': ["detect"],
                'active': True,
                'activationTime': datetime.datetime.utcnow()
        }).inserted_id
        time.sleep(1.5)
        self.assertFalse(sch.scheduleDetection('test', 'test_conf', []))
        jServer.delete_one({'_id': ObjectId(jID)})
        ioc.getCollection('SourceData').drop()

    def test_detectionScheduling(self):
        ioc = GreaseContainer()
        ioc.ensureRegistration()
        sch = Scheduling(ioc)
        jServer = ioc.getCollection('JobServer')
        jID1 = jServer.insert_one({
                'jobs': 0,
                'os': platform.system().lower(),
                'roles': ["general"],
                'prototypes': ["detect"],
                'active': True,
                'activationTime': datetime.datetime.utcnow()
        }).inserted_id
        time.sleep(1)
        jID2 = jServer.insert_one({
                'jobs': 0,
                'os': platform.system().lower(),
                'roles': ["general"],
                'prototypes': ["detect"],
                'active': True,
                'activationTime': datetime.datetime.utcnow()
        }).inserted_id
        time.sleep(1)
        self.assertTrue(sch.scheduleDetection('test', 'test_conf', [
            {
                'test0': 'var0',
                'test1': 'var1',
                'test2': 'var2',
                'test3': 'var3',
                'test4': 'var4',
                'test5': 'var5',
                'test6': 'var6',
                'test7': 'var7',
                'test8': 'var8',
                'test9': 'var9',
                'test10': 'var10',
            },
            {
                'test0': 'var0',
                'test1': 'var1',
                'test2': 'var2',
                'test3': 'var3',
                'test4': 'var4',
                'test5': 'var5',
                'test6': 'var6',
                'test7': 'var7',
                'test8': 'var8',
                'test9': 'var9',
                'test10': 'var10',
            },
            {
                'test0': 'var0',
                'test1': 'var1',
                'test2': 'var2',
                'test3': 'var3',
                'test4': 'var4',
                'test5': 'var5',
                'test6': 'var6',
                'test7': 'var7',
                'test8': 'var8',
                'test9': 'var9',
                'test10': 'var10',
            },
            {
                'test0': 'var0',
                'test1': 'var1',
                'test2': 'var2',
                'test3': 'var3',
                'test4': 'var4',
                'test5': 'var5',
                'test6': 'var6',
                'test7': 'var7',
                'test8': 'var8',
                'test9': 'var9',
                'test10': 'var10',
            },
            {
                'test0': 'var0',
                'test1': 'var1',
                'test2': 'var2',
                'test3': 'var3',
                'test4': 'var4',
                'test5': 'var5',
                'test6': 'var6',
                'test7': 'var7',
                'test8': 'var8',
                'test9': 'var9',
                'test10': 'var10',
            },
            {
                'test0': 'var0',
                'test1': 'var1',
                'test2': 'var2',
                'test3': 'var3',
                'test4': 'var4',
                'test5': 'var5',
                'test6': 'var6',
                'test7': 'var7',
                'test8': 'var8',
                'test9': 'var9',
                'test10': 'var10',
            },
        ]))
        time.sleep(1)
        self.assertEqual(ioc.getCollection('SourceData').find({
            'grease_data.detection.server': ObjectId(jID1)
        }).count(), 3)
        self.assertEqual(ioc.getCollection('SourceData').find({
            'grease_data.detection.server': ObjectId(jID2)
        }).count(), 3)
        self.assertEqual(ioc.getCollection('JobServer').find_one({
            '_id': ObjectId(jID1)
        })['jobs'], 3)
        self.assertEqual(ioc.getCollection('JobServer').find_one({
            '_id': ObjectId(jID2)
        })['jobs'], 3)
        jServer.delete_one({'_id': ObjectId(jID1)})
        jServer.delete_one({'_id': ObjectId(jID2)})
        ioc.getCollection('SourceData').drop()

    def test_scheduleScheduling(self):
        d = Detect()
        p = PrototypeConfig(d.ioc)
        source = {
            'key': 'var',
            'ver': 'key',
            'greg': 'old',
            'created': '2017-11-24'
        }
        configuration = {
            'name': 'demoConfig',
            'job': 'otherThing',
            'exe_env': 'general',
            'source': 'Google',
            'logic': {
                'DateRange': [
                    {
                        'field': 'created',
                        'format': '%Y-%m-%d',
                        'min': '2017-11-23',
                        'max': '2017-11-25'
                    }
                ],
                'Regex': [
                    {
                        'field': 'key',
                        'pattern': '.*',
                        'variable': True,
                        'variable_name': 'field'
                    },
                    {
                        'field': 'ver',
                        'pattern': '.*'
                    }
                ],
                'Exists': [
                    {
                        'field': 'greg',
                        'variable': True,
                        'variable_name': 'greg'
                    }
                ]
            }
        }
        p.load(True, [configuration])
        sourceId = d.ioc.getCollection('SourceData').insert_one({
                    'grease_data': {
                        'sourcing': {
                            'server': ObjectId(d.ioc.getConfig().NodeIdentity)
                        },
                        'detection': {
                            'server': ObjectId(d.ioc.getConfig().NodeIdentity),
                            'detectionStart': None,
                            'detectionEnd': None,
                            'detection': {}
                        },
                        'scheduling': {
                            'schedulingServer': None,
                            'schedulingStart': None,
                            'schedulingEnd': None
                        },
                        'execution': {
                            'server': None,
                            'assignmentTime': None,
                            'executionStart': None,
                            'executionEnd': None,
                            'context': {},
                            'executionSuccess': False,
                            'commandSuccess': False,
                            'failures': 0,
                            'retryTime': datetime.datetime.utcnow()
                        }
                    },
                    'source': str('test').encode('utf-8'),
                    'configuration': configuration.get('name'),
                    'data': source,
                    'createTime': datetime.datetime.utcnow(),
                    'expiry': Deduplication.generate_max_expiry_time(1)
        }).inserted_id
        scheduleServer = d.ioc.getCollection('JobServer').insert_one({
                'jobs': 0,
                'os': platform.system().lower(),
                'roles': ["general"],
                'prototypes': ["schedule"],
                'active': True,
                'activationTime': datetime.datetime.utcnow()
        }).inserted_id
        self.assertTrue(d.detectSource())
        self.assertFalse(d.getScheduledSource())
        self.assertTrue(len(
            d.ioc.getCollection('SourceData').find_one(
                {
                    'grease_data.scheduling.server': ObjectId(scheduleServer),
                    'grease_data.scheduling.start': None,
                    'grease_data.scheduling.end': None
                },
                sort=[('grease_data.createTime', pymongo.DESCENDING)]
            )
        ))
        d.ioc.getCollection('JobServer').delete_one({'_id': ObjectId(scheduleServer)})
        d.ioc.getCollection('SourceData').delete_one({'_id': ObjectId(sourceId)})



from unittest import TestCase
from tgt_grease.enterprise.Detectors import DateRange


class TestDateRange(TestCase):

    def test_range_compare_min_pass(self):
        r = DateRange()
        self.assertTrue(r.timeCompare('2017-11-24', {'min': '2017-11-24', 'format': '%Y-%m-%d'}))

    def test_range_compare_min_fail(self):
        r = DateRange()
        self.assertFalse(r.timeCompare('2017-11-24', {'min': '2017-11-25', 'format': '%Y-%m-%d'}))

    def test_range_compare_max_pass(self):
        r = DateRange()
        self.assertTrue(r.timeCompare('2017-11-24', {'max': '2017-11-25', 'format': '%Y-%m-%d'}))

    def test_range_compare_max_fail(self):
        r = DateRange()
        self.assertFalse(r.timeCompare('2017-11-24', {'max': '2017-11-23', 'format': '%Y-%m-%d'}))

    def test_range_compare_min_and_max_pass(self):
        r = DateRange()
        self.assertTrue(r.timeCompare('2017-11-25', {'min': '2017-11-24', 'max': '2017-11-26', 'format': '%Y-%m-%d'}))

    def test_range_compare_min_and_max_fail(self):
        r = DateRange()
        self.assertFalse(r.timeCompare('2017-11-32', {'min': '2017-11-24', 'max': '2017-11-26', 'format': '%Y-%m-%d'}))

    def test_range_compare_no_min_or_max(self):
        r = DateRange()
        self.assertFalse(r.timeCompare('2017-11-24', {}))

    def test_range_compare_bad_field_type(self):
        r = DateRange()
        self.assertFalse(r.timeCompare('cats', {'min': '2017-11-24', 'max': '2017-11-26', 'format': '%Y-%m-%d'}))

    def test_range_compare_bad_min_or_max(self):
        r = DateRange()
        self.assertFalse(r.timeCompare('2017-11-23', {'min': '2017-11-24', 'format': '%Y-%m-%d'}))
        self.assertFalse(r.timeCompare('2017-11-25', {'max': '2017-11-24', 'format': '%Y-%m-%d'}))
        self.assertFalse(r.timeCompare('2017-11-27', {'min': '2017-11-24', 'max': '2017-11-26', 'format': '%Y-%m-%d'}))

    def test_emptyList(self):
        r = DateRange()
        state, data = r.processObject({}, [])
        self.assertFalse(state)
        self.assertFalse(data)

    def test_valid(self):
        r = DateRange()
        state, data = r.processObject(
            {
                'test': u'data',
                'ver': 'var',
                'data': '2017-11-24'
            },
            [
                {
                    'field': 'data',
                    'min': '2017-11-24',
                    'format': '%Y-%m-%d',
                    'variable': True,
                    'variable_name': 'data'
                }
            ]
        )
        self.assertTrue(state)
        self.assertEqual(data.get('data'), '2017-11-24')

    def test_valid_no_variables(self):
        r = DateRange()
        state, data = r.processObject(
            {
                'test': u'data',
                'ver': 'var',
                'data': '2017-11-24'
            },
            [
                {
                    'field': 'data',
                    'min': '2017-11-24',
                    'format': '%Y-%m-%d'
                }
            ]
        )
        self.assertTrue(state)
        self.assertEqual(len(data), 0)

    def test_invalid_config(self):
        r = DateRange()
        state, data = r.processObject(
            {
                'test': u'data',
                'ver': 'var',
                'data': '2017-11-24'
            },
            [
                {
                    'field': 'data',
                    'min': '2017-11-24'
                    # no format
                }
            ]
        )
        self.assertFalse(state)
        self.assertEqual(len(data), 0)

    def test_bad_source(self):
        r = DateRange()
        state, data = r.processObject(
            [],
            [
                {
                    'field': 'data',
                    'min': '2017-11-24',
                    'format': '%Y-%m-%d',
                    'variable': True,
                    'variable_name': 'data'
                }
            ]
        )
        self.assertFalse(state)
        self.assertEqual(len(data), 0)

    def test_bad_config(self):
        r = DateRange()
        state, data = r.processObject(
            {
                'test': u'data',
                'ver': 'var',
                'data': '2017-11-24'
            },
            {}
        )
        self.assertFalse(state)
        self.assertEqual(len(data), 0)

    def test_empty_result(self):
        r = DateRange()
        state, data = r.processObject(
            {
                'test': u'data',
                'ver': 'var',
                'data': '2017-11-23'
            },
            [
                {
                    'field': 'data',
                    'min': '2017-11-24',
                    'format': '%Y-%m-%d',
                    'variable': True,
                    'variable_name': 'data'
                }
            ]
        )
        self.assertFalse(state)
        self.assertEqual(len(data), 0)

    def test_field_not_exist(self):
        r = DateRange()
        state, data = r.processObject(
            {
                'test': u'data',
                'ver': 'var',
                'data': '2017-11-24'
            },
            [
                {
                    'field': 'data1',
                    'min': '2017-11-24',
                    'format': '%Y-%m-%d',
                    'variable': True,
                    'variable_name': 'data'
                }
            ]
        )
        self.assertFalse(state)
        self.assertEqual(len(data), 0)

    def test_field_false(self):
        r = DateRange()
        state, data = r.processObject(
            {
                'test': u'data',
                'ver': 'var',
                'data': False
            },
            [
                {
                    'field': 'data',
                    'min': '2017-11-24',
                    'format': '%Y-%m-%d',
                    'variable': True,
                    'variable_name': 'data'
                }
            ]
        )
        self.assertFalse(state)
        self.assertEqual(len(data), 0)

    def test_bad_block(self):
        r = DateRange()
        state, data = r.processObject(
            {
                'test': u'data',
                'ver': 'var',
                'data': '2017-11-24'
            },
            [
                []
            ]
        )
        self.assertFalse(state)
        self.assertEqual(len(data), 0)



from unittest import TestCase
from tgt_grease.core import ImportTool, Configuration, Logging


class TestImporter(TestCase):
    def test_load(self):
        log = Logging()
        imp = ImportTool(log)
        Conf = imp.load("Configuration")
        self.assertTrue(isinstance(Conf, Configuration))

    def test_failed_path(self):
        log = Logging()
        imp = ImportTool(log)
        obj = imp.load("defaultdict")
        self.assertFalse(obj)



from tgt_grease.enterprise.Model import Detector


class Range(Detector):
    """Field Number Range Detector for GREASE Detection

    A Typical Range configuration looks like this::

        {
            ...
            'logic': {
                'Range': [
                    {
                        'field': String, # <-- Field to search for
                        'min': Int/float, # <-- OPTIONAL IF max is set
                        'max': Int/Float # <-- OPTIONAL IF min is set
                    }
                    ...
                ]
                ...
            }
        }

    Note:
        The min field is only required if max is not present. This would ensure the field is above the number provided
    Note:
        The max field is only required if min is not present. This would ensure the field is below the number provided
    Note:
        If both are provided then the field would need to be inside the range provided
    Note:
        To find an exact number set min one below the target and max one above the target

    """

    def processObject(self, source, ruleConfig):
        """Processes an object and returns valid rule data

        Data returned in the second parameter from this method should be in this form::

            {
                '<field>': Object # <-- if specified as a variable then return the key->Value pairs
                ...
            }

        Args:
            source (dict): Source Data
            ruleConfig (list[dict]): Rule Configuration Data

        Return:
            tuple: first element boolean for success; second dict for any fields returned as variables

        """
        final = {}
        finalBool = False
        # type checks
        if not isinstance(source, dict):
            return False, {}
        if not isinstance(ruleConfig, list):
            return False, {}
        else:
            # loop through configuration for each set of logical configurations
            for block in ruleConfig:
                if not isinstance(block, dict):
                    self.ioc.getLogger().error(
                        "INVALID RANGE LOGICAL BLOCK! NOT TYPE LIST [{0}]".format(str(type(block))),
                        notify=False
                    )
                    return False, {}
                else:
                    # look for field and perform range check
                    if block.get('field') in source:
                        # perform range operation since field exists
                        if source.get(block.get('field')):
                            if self.range_compare(source.get(block.get('field')), block):
                                finalBool = True
                                if block.get('variable') and block.get('variable_name'):
                                    final[str(block.get('variable_name'))] = source.get(block.get('field'))
                                else:
                                    continue
                            else:
                                self.ioc.getLogger().trace("Field Failed Range Comparison", verbose=True)
                                return False, {}
                        else:
                            # truthy false field value
                            self.ioc.getLogger().trace(
                                "Field [{0}] equated to false [{1}]".format(
                                    block.get('field'),
                                    source.get(block.get('field'))
                                ),
                                verbose=True
                            )
                            return False, {}
                    else:
                        self.ioc.getLogger().trace(
                            "Field not found in source [{0}]".format(block.get('field')),
                            verbose=True
                        )
                        return False, {}
        return finalBool, final

    def range_compare(self, field, LogicalBlock):
        """Compares number range to a field

        Args:
            field (int/float/long): field to compare
            LogicalBlock (dict): Logical Block

        Returns:
            bool: if the range is successful then true else false

        """
        # Ensure field is present
        # ensure at least min OR max is present
        if not LogicalBlock.get('min') and not LogicalBlock.get('max'):
            self.ioc.getLogger().trace("[min] and/or [max] not found in config block", verbose=True)
            return False
        if LogicalBlock.get('min') and not isinstance(LogicalBlock.get('min'), (int, float)):
            self.ioc.getLogger().trace("min not of type int or float", verbose=True)
            return False
        if LogicalBlock.get('max') and not isinstance(LogicalBlock.get('max'), (int, float)):
            self.ioc.getLogger().trace("max not of type int or float", verbose=True)
            return False
        # ensure field is an number
        if not isinstance(field, (int, float)):
            self.ioc.getLogger().trace("Field is NaN", verbose=True)
            return False
        if LogicalBlock.get('min') and LogicalBlock.get('max'):
            # Min & Max Defined
            if isinstance(field, int):
                minValue = int(LogicalBlock.get('min'))
                maxValue = int(LogicalBlock.get('max'))
                field = int(field)
            elif isinstance(field, float):
                minValue = float(LogicalBlock.get('min'))
                maxValue = float(LogicalBlock.get('max'))
                field = float(field)
            else:
                self.ioc.getLogger().error("Internal Casting Error; Failed to determine field type", notify=False)
                return False
            if minValue <= field <= maxValue:
                return True
            else:
                return False
        elif LogicalBlock.get('min') and not LogicalBlock.get('max'):
            # Min Defined
            if isinstance(field, int):
                minValue = int(LogicalBlock.get('min'))
                field = int(field)
            elif isinstance(field, float):
                minValue = float(LogicalBlock.get('min'))
                field = float(field)
            else:
                self.ioc.getLogger().error("Internal Casting Error; Failed to determine field type", notify=False)
                return False
            if field >= minValue:
                return True
            else:
                return False
        elif not LogicalBlock.get('min') and LogicalBlock.get('max'):
            # Max Defined
            if isinstance(field, int):
                maxValue = int(LogicalBlock.get('max'))
                field = int(field)
            elif isinstance(field, float):
                maxValue = float(LogicalBlock.get('max'))
                field = float(field)
            else:
                self.ioc.getLogger().error("Internal Casting Error; Failed to determine field type", notify=False)
                return False
            if field <= maxValue:
                return True
            else:
                return False
        else:
            self.ioc.getLogger().error("Failed to find either min OR max in LogicalBlock", verbose=True, notify=False)
            return False



from abc import ABCMeta, abstractmethod


class BaseSourceClass(object):
    """Base Class for all sources to implement

    Attributes:
        _data (list[dict]): List of data to be returned to GREASE
        deduplication_strength (float): Level of deduplication strength to use *higher is stronger uniqueness*
        field_set (None or list): If none all fields found will be duplicated otherwise only fields listed will be
        deduplication_expiry (int): Hours to retain deduplication data
        deduplication_expiry_max (int): Days to deduplicate for **maximum**

    """

    __metaclass__ = ABCMeta

    def __init__(self):
        self._data = []
        self.deduplication_strength = 85.0
        self.deduplication_expiry = 12
        self.deduplication_expiry_max = 7
        self.field_set = None

    @abstractmethod
    def mock_data(self, configuration):
        """Mock the source for data

        Use this method to read through configuration provided to you, and mock getting data. This will *always* be
        called by the scan engine. **Ensure you set any data to the `self._data` variable. A list of dictionaries for
        the engine to schedule for detection**

        Args:
            configuration (dict): Configuration for the sourcing to occur with

        Note:
            This is the method to fill out to get data into GREASE.

        Returns:
            list[dict]: mock data from source

        """
        pass

    @abstractmethod
    def parse_source(self, configuration):
        """Parse the source for data

        Use this method to read through configuration provided to you, and get data. This will *always* be called
        by the scan engine. **Ensure you set any data to the `self._data` variable. A list of dictionaries for the
        engine to schedule for detection**

        Args:
            configuration (dict): Configuration for the sourcing to occur with

        Note:
            This is the method to fill out to get data into GREASE.

        Returns:
            bool: If True data will be scheduled for ingestion after deduplication. If False the engine will bail out

        """
        pass

    def get_data(self):
        """Returns data from source

        Returns:
            list[dict]: List of *single dimension* dictionaries for GREASE to parse through other prototypes

        """
        return self._data



from tgt_grease.core import Configuration, Notifications
from logging import config
from datetime import datetime
import logging
import os
import time

GREASE_LOG_HANDLER = None


class Logging(object):
    """Application Logging for GREASE

    This is the primary configuration source for GREASE logging. All log information will be passed here to
    enable centralized log aggregation

    Attributes:
        _conf (Configuration): This is an instance of the Config to enable configuring loggers
        _logger (logging.Logger): This is the actual logger for GREASE
        _formatter (logging.Formatter): This is the log formatter
        _notifications (Notifications): Notifications instance
        foreground (bool): If set will override config and print all log messages

    """

    _conf = None
    _logger = None
    _formatter = None
    _notifications = None
    foreground = False

    def __init__(self, Config=None):
        if isinstance(Config, Configuration):
            self._conf = Config
        else:
            self._conf = Configuration()
        self._notifications = Notifications(self.getConfig())
        self.ProvisionLoggers()

    def getConfig(self):
        """Getter for Configuration

        Returns:
            Configuration: The loaded configuration object

        """
        return self._conf

    def getNotification(self):
        """Get Notification Class

        Returns:
            Notifications: The current Notifications instance

        """
        return self._notifications

    def TriageMessage(self, message, additional=None, verbose=False, trace=False, notify=False, level=logging.DEBUG):
        """Central message handler

        Args:
            message (str): Message to Log
            additional (object): Additional information to log
            verbose (bool): To be printed if verbose is enabled
            trace (bool): To be printed if trace is enabled
            notify (bool): If true will pass through notification system
            level (int): Log Level

        Returns:
            bool: Log Success

        """
        # first prevent verbose processing
        if verbose and not self._conf.get('Logging', 'verbose'):
            return True
        # prevent trace processing
        if trace and not self._conf.get('Logging', 'trace'):
            return True
        # create a pre-message
        if level is 0:
            preMsg = "TRACE"
        elif level is logging.DEBUG:
            preMsg = "DEBUG"
        elif level is logging.INFO:
            preMsg = "INFO"
        elif level is logging.WARNING:
            preMsg = "WARNING"
        elif level is logging.ERROR:
            preMsg = "ERROR"
        elif level is logging.CRITICAL:
            preMsg = "CRITICAL"
        else:
            preMsg = "UNSET"
        if verbose:
            preMsg = "VERBOSE::{0}".format(preMsg)
        if trace:
            preMsg = "TRACE::{0}".format(preMsg)
        if additional:
            message = "{0}::{1}::{2}::{3}".format(preMsg, self._conf.NodeIdentity, message, additional)
        else:
            message = "{0}::{1}::{2}".format(preMsg, self._conf.NodeIdentity, message)
        # Foreground mode print log messages
        if self._conf.get('Logging', 'foreground') or self.foreground:
            print("{0}::{1}".format(datetime.utcnow(), message))
        # actually log the message
        if level is 0:
            self._logger.log(logging.DEBUG, message)
        else:
            self._logger.log(level, message)
        # notify if needed
        if notify:
            return bool(self._notifications.SendMessage(message, level))
        return True

    def trace(self, message, additional=None, verbose=False, trace=True, notify=False):
        """Trace Messages

        Use this method for logging tracing (enhanced debug) statements

        Args:
            message (str): Message to log
            additional (object): Additional information to log. Note: object must be able to transform to string
            verbose (bool): Print only if verbose mode
            trace (bool): Print only if trace mode
            notify (bool): Run through the notification management system

        Returns:
            bool: Message is logged

        """
        return bool(self.TriageMessage(
            message,
            additional=additional,
            verbose=verbose,
            trace=trace,
            notify=notify,
            level=0
        ))

    def debug(self, message, additional=None, verbose=False, trace=False, notify=False):
        """Debug Messages

        Use this method for logging debug statements

        Args:
            message (str): Message to log
            additional (object): Additional information to log. Note: object must be able to transform to string
            verbose (bool): Print only if verbose mode
            trace (bool): Print only if trace mode
            notify (bool): Run through the notification management system

        Returns:
            bool: Message is logged

        """
        return bool(self.TriageMessage(
            message,
            additional=additional,
            verbose=verbose,
            trace=trace,
            notify=notify,
            level=logging.DEBUG
        ))

    def info(self, message, additional=None, verbose=False, trace=False, notify=False):
        """Info Messages

        Use this method for logging info statements

        Args:
            message (str): Message to log
            additional (object): Additional information to log. Note: object must be able to transform to string
            verbose (bool): Print only if verbose mode
            trace (bool): Print only if trace mode
            notify (bool): Run through the notification management system

        Returns:
            bool: Message is logged

        """
        return bool(self.TriageMessage(
            message,
            additional=additional,
            verbose=verbose,
            trace=trace,
            notify=notify,
            level=logging.INFO
        ))

    def warning(self, message, additional=None, verbose=False, trace=False, notify=False):
        """Warning Messages

        Use this method for logging warning statements

        Args:
            message (str): Message to log
            additional (object): Additional information to log. Note: object must be able to transform to string
            verbose (bool): Print only if verbose mode
            trace (bool): Print only if trace mode
            notify (bool): Run through the notification management system

        Returns:
            bool: Message is logged

        """
        return bool(self.TriageMessage(
            message,
            additional=additional,
            verbose=verbose,
            trace=trace,
            notify=notify,
            level=logging.WARNING
        ))

    def error(self, message, additional=None, verbose=False, trace=False, notify=True):
        """Error Messages

        Use this method for logging error statements

        Args:
            message (str): Message to log
            additional (object): Additional information to log. Note: object must be able to transform to string
            verbose (bool): Print only if verbose mode
            trace (bool): Print only if trace mode
            notify (bool): Run through the notification management system

        Returns:
            bool: Message is logged

        """
        return bool(self.TriageMessage(
            message,
            additional=additional,
            verbose=verbose,
            trace=trace,
            notify=notify,
            level=logging.ERROR
        ))

    def critical(self, message, additional=None, verbose=False, trace=False, notify=True):
        """Critical Messages

        Use this method for logging critical statements

        Args:
            message (str): Message to log
            additional (object): Additional information to log. Note: object must be able to transform to string
            verbose (bool): Print only if verbose mode
            trace (bool): Print only if trace mode
            notify (bool): Run through the notification management system

        Returns:
            bool: Message is logged

        """
        return bool(self.TriageMessage(
            message,
            additional=additional,
            verbose=verbose,
            trace=trace,
            notify=notify,
            level=logging.CRITICAL
        ))

    def ProvisionLoggers(self):
        """Loads Log Handler & Config

        Returns:
            None: Simple loader, Nothing needed

        """
        if self._conf.get('Logging', 'ConfigurationFile'):
            if os.path.isfile(self._conf.get('Logging', 'ConfigurationFile')):
                config.fileConfig(self._conf.get('Logging', 'ConfigurationFile'))
                self._logger = logging.getLogger('GREASE')
            else:
                self.DefaultLogger()
        else:
            self.DefaultLogger()

    def DefaultLogger(self):
        """Default Logging Provisioning

        Returns:
            None: void method to provision class internals

        """
        global GREASE_LOG_HANDLER
        self._logger = logging.getLogger('GREASE')
        self._logger.setLevel(logging.DEBUG)
        self._formatter = logging.Formatter(
            "{"
            "\"timestamp\": \"%(asctime)s.%(msecs)03d\", "
            "\"thread\": \"%(threadName)s\", "
            "\"level\" : \"%(levelname)s\", "
            "\"message\" : \"%(message)s\"}",
            "%Y-%m-%d %H:%M:%S"
        )
        self._formatter.converter = time.gmtime
        if not GREASE_LOG_HANDLER:
            if os.path.isdir(self._conf.greaseDir):
                GREASE_LOG_HANDLER = logging.FileHandler(self._conf.get('Logging', 'file'))
                GREASE_LOG_HANDLER.setLevel(logging.DEBUG)
                GREASE_LOG_HANDLER.setFormatter(self._formatter)
                self._logger.addHandler(GREASE_LOG_HANDLER)
            else:
                GREASE_LOG_HANDLER = logging.StreamHandler()
                GREASE_LOG_HANDLER.setLevel(logging.DEBUG)
                GREASE_LOG_HANDLER.setFormatter(self._formatter)
                self._logger.addHandler(GREASE_LOG_HANDLER)



from tgt_grease.core import GreaseContainer
from psutil import virtual_memory, cpu_percent
from bson.objectid import ObjectId
import threading
import hashlib
import datetime
import difflib
import pymongo


class Deduplication(object):
    """Responsible for Deduplication Operations

    Deduplication in GREASE is a multi-step process to ensure performance and accuracy of deduplication. The overview of
    this process is this:
        - Step 1: Identify a Object Type 1 Hash Match. A Type 1 Object (T1) is a SHA256 hash of a dictionary in a data list. If we can hash the entire object and find a match then the object is 100% duplicate.
        - Step 2: Object Type 2 Matching. If a Type 1 (T1) object cannot be found Type 2 Object (T2) deduplication occurs. This will introspect the dictionary for each field and map them against other likely objects of the same type. If a hash match is found (source + field + value as a SHA256) then the field is 100% duplicate. The aggregate score of all fields or the specified subset is above the provided threshold then the object is duplicate. This prevents similar objects from passing through when they are most likely updates to an original object that does not need to be computed on. If a field updates that you will need always then exclude it will need to be passed into the `Deduplicate` function.

    Object examples::

        # Type 1 Object

        {
            '_id': ObjectId, # <-- MongoDB ObjectID
            'type: Int, # <-- Always Type 1
            'hash': String, # <-- SHA256 hash of entire object
            'expiry': DateTime, # <-- Expiration time if no objects are found to be duplicate after which object will be deleted
            'max_expiry': DateTime, # <-- Expiration time for object to be deleted when reached
            'score': Int, # <-- Amount of times this object has been found
            'source': String # <-- Source of the object
        }
        # Type 2 Object
        {
            '_id': ObjectId, # <-- MongoDB ObjectID
            'type: Int, # <-- Always Type 2
            'source': String, # <-- Source of data
            'field': String, # <-- Field in Object
            'value': String, # <-- Value of Object's field
            'hash': String, # <-- SHA256 of source + field + value
            'expiry': DateTime, # <-- Expiration time if no objects are found to be duplicate after which object will be deleted
            'max_expiry': DateTime, # <-- Expiration time for object to be deleted when reached
            'score': Int, # <-- Amount of times this object has been found
            'parentId': ObjectId # <-- T1 Object ID from parent
        }

    Attributes:
        ioc (GreaseContainer): IoC access for DeDuplication

    """

    def __init__(self, ioc=None):
        if isinstance(ioc, GreaseContainer):
            self.ioc = ioc
        else:
            self.ioc = GreaseContainer()

    def Deduplicate(self, data, source, configuration, threshold, expiry_hours, expiry_max, collection, field_set=None):
        """Deduplicate data

        This method will deduplicate the `data` object to allow for only unique objects to be returned. The collection
        variable will be the collection deduplication data will be stored in

        Args:
            data (list[dict]): **list or single dimensional dictionaries** to deduplicate
            source (str): Source of data being deduplicated
            configuration (str): Configuration Name Provided
            threshold (float): level of duplication allowed in an object (the lower the threshold the more uniqueness is required)
            expiry_hours (int): Hours to retain deduplication data
            expiry_max (int): Maximum days to retain deduplication data
            collection (str): Deduplication collection to use
            field_set (list, optional): Fields to deduplicate on

        Note:
            expiry_hours is specific to how many hours objects will be persisted for if they are not seen again

        Returns:
            list[dict]: Deduplicated data

        """
        # ensure we got a list
        if not isinstance(data, list):
            self.ioc.getLogger().error(
                "Data was not of type list for Deduplication got type [{0}]".format(str(type(data))),
                notify=False,
                verbose=True
            )
            return []
        # ensure there is data to parse
        if len(data) <= 0:
            # empty list return empty lists
            return []
        self.ioc.getLogger().trace(
            "Starting deduplication from data source [{0}] total records to parse [{1}]".format(source, len(data)),
            trace=True
        )
        # now comes James' version of machine learning. I call it "Blue Collar Machine Learning"
        # Pointer to access items in the list
        data_pointer = 0
        # Max Length
        data_max = len(data)
        if data_max is 0:
            # we have no data to process
            self.ioc.getLogger().trace("Length of data is zero", verbose=True)
            return []
        # Thread pool
        threads = []
        # Final result
        final = []
        # loop through the objects
        while data_pointer < data_max:
            # ensure we don't swamp the system resources
            cpu = cpu_percent(interval=.1)
            mem = virtual_memory().percent
            if \
                    cpu >= int(self.ioc.getConfig().get('NodeInformation', 'ResourceMax')) or \
                    mem >= int(self.ioc.getConfig().get('NodeInformation', 'ResourceMax')):
                self.ioc.getLogger().trace("Deduplication sleeping; System resource maximum reached", verbose=True)
                # remove variables
                del cpu
                del mem
                continue
            # Resources are available lets start cooking down this list
            # Poll the active threads to ensure we are cleaning up
            self.ioc.getLogger().trace("Thread Pool polling Starting", verbose=True)
            threads_final = []
            for thread in threads:
                if thread.isAlive():
                    threads_final.append(thread)
            threads = threads_final
            self.ioc.getLogger().trace("Thread polling complete", verbose=True)
            self.ioc.getLogger().trace("Total current deduplication threads [{0}]".format(len(threads)), verbose=True)
            # ensure we do not breach the thread limit for the server
            if len(threads) >= int(self.ioc.getConfig().get('NodeInformation', 'DeduplicationThreads')):
                self.ioc.getLogger().trace(
                    "Thread max reached. Deduplication waiting for threads to complete",
                    verbose=True
                )
                continue
            # Ensure each object is a dictionary
            if not isinstance(data[data_pointer], dict):
                self.ioc.getLogger().warning(
                    'DeDuplication Received NON-DICT from source: [{0}] Type: [{1}] got: [{2}]'.format(
                        source,
                        str(type(data[data_pointer])),
                        str(data[data_pointer])
                    )
                )
                data_pointer += 1
                continue
            # create thread for deduplication
            proc = threading.Thread(
                target=self.deduplicate_object,
                args=(
                    self.ioc,
                    data[data_pointer],
                    expiry_hours,
                    expiry_max,
                    threshold,
                    source,
                    configuration,
                    final,
                    collection,
                    data_pointer,
                    data_max,
                    field_set,
                ),
                name="GREASE DEDUPLICATION THREAD [{0}/{1}]".format(data_pointer, data_max)
            )
            proc.daemon = True
            proc.start()
            threads.append(proc)
            data_pointer += 1
            self.ioc.getLogger().trace("Total current deduplication threads [{0}]".format(len(threads)), verbose=True)
        self.ioc.getLogger().info("All data objects have been threaded for processing", verbose=True)
        # wait for threads to finish out
        while len(threads) > 0:
            self.ioc.getLogger().trace("Total current deduplication threads [{0}]".format(len(threads)), verbose=True)
            threads_final = []
            for thread in threads:
                if thread.isAlive():
                    threads_final.append(thread)
            threads = threads_final
            self.ioc.getLogger().trace("Total current deduplication threads [{0}]".format(len(threads)), verbose=True)
        # ensure collections expiry timers are in place
        self.ioc.getCollection(collection).create_index([('expiry', 1), ('expireAfterSeconds', 1)])
        self.ioc.getCollection(collection).create_index([('max_expiry', 1), ('expireAfterSeconds', 1)])
        return final

    @staticmethod
    def deduplicate_object(ioc, obj, expiry, expiry_max, threshold, source_name, configuration_name, final, collection, data_pointer=None, data_max=None, field_set=None):
        """DeDuplicate Object

        This is the method to actually deduplicate an object. The `final` argument is appended to with the obj if it
        was successfully deduplicated.

        Args:
            ioc (GreaseContainer): IoC for the instance
            obj (dict): Object to be deduplicated
            expiry (int): Hours to deduplicate for
            expiry_max (int): Maximum days to deduplicate for
            threshold (float): level of duplication allowed in an object (the lower the threshold the more uniqueness is required)
            source_name (str): Source of data being deduplicated
            configuration_name (str): Configuration being deduplicated for
            final (list): List to append `obj` to if unique
            collection (str): Name of deduplication collection
            data_pointer (int): If provided will provide log information relating to thread
                (Typically used via `Deduplicate`)
            data_max (int): If provided will provide log information relating to thread
                (Typically used via `Deduplicate`)
            field_set (list): If provided will only deduplicate on list of fields provided

        Returns:
            None: Nothing returned. Updates `final` object

        """
        # first determine if this object has been seen before
        DeDupCollection = ioc.getCollection(collection)
        t1test = obj
        t1test['grease_internal_configuration'] = configuration_name
        T1Hash = DeDupCollection.find_one({'hash': Deduplication.generate_hash_from_obj(t1test)})
        if T1Hash:
            # T1 Found Protocol: We have found a fully duplicate object
            # we have a duplicate source document
            # increase the counter and expiry and move on (DROP)
            ioc.getLogger().debug("Type1 Match found for object", verbose=True)
            # bump the expiry time and move on
            DeDupCollection.update_one(
                {'_id': T1Hash['_id']},
                {
                    "$set": {
                        'score': int(T1Hash['score']) + 1,
                        'expiry': Deduplication.generate_expiry_time(expiry)
                    }
                }
            )
            return
        else:
            # T1 Not Found Protocol: We have a possibly unique object
            ioc.getLogger().debug("Type1 Match not found; Beginning type 2 processing")
            # Create a T1
            T1ObjectId = DeDupCollection.insert_one({
                'expiry': Deduplication.generate_expiry_time(int(expiry)),
                'grease_internal_configuration': configuration_name,
                'max_expiry': Deduplication.generate_max_expiry_time(int(expiry_max)),
                'type': 1,
                'score': 1,
                'source': str(source_name),
                'hash': Deduplication.generate_hash_from_obj(t1test)
            }).inserted_id
            # Begin T2 Deduplication
            compositeScore = Deduplication.object_field_score(
                collection, ioc, source_name, configuration_name, obj, str(T1ObjectId), expiry, expiry_max, field_set
            )
            if compositeScore < threshold:
                # unique obj
                ioc.getLogger().trace(
                    "Unique object! Composite score was: [{0}] threashold: [{1}]".format(compositeScore, threshold),
                    verbose=True
                )
                final.append(obj)
                return
            else:
                # likely duplicate value
                ioc.getLogger().trace(
                    "Object surpassed threshold, suspected to be duplicate! "
                    "Composite score was: [{0}] threashold: [{1}]".format(compositeScore, threshold),
                    verbose=True
                )
                return

    @staticmethod
    def object_field_score(collection, ioc, source_name, configuration_name, obj, objectId, expiry, max_expiry, field_set=None):
        """Returns T2 average uniqueness

        Takes a dictionary and returns the likelihood of that object being unique based on data in the collection

        Args:
            collection (str): Deduplication collection name
            ioc (GreaseContainer): IoC Access
            source_name (str): source of data to be deduplicated
            configuration_name (str): configuration name to be deduplicated
            obj (dict): Single dimensional list to be compared against collection
            objectId (str): T1 Hash Mongo ObjectId to be used to associate fields to a T1
            expiry (int): Hours for deduplication to wait before removing a field if not seen again
            max_expiry (int): Days for deduplication to wait before ensuring object is deleted
            field_set (list, optional): List of fields to deduplicate with if provided. Else will use all keys

        Returns:
            float: Duplication Probability

        """
        # generate field list if not provided
        FieldColl = ioc.getCollection(collection)
        if not isinstance(field_set, list) or len(field_set) <= 0:
            field_set = obj.keys()
        # List to hold field level scores
        field_scores = []
        # iterate over the field set
        for field in field_set:
            # ensure key is in the object
            ioc.getLogger().trace("Starting field [{0}]".format(field), verbose=True)
            if field in obj:
                if isinstance(obj.get(field), bytes):
                    value = obj.get(field).decode('utf-8', 'ignore')
                else:
                    value = obj.get(field)
                T2Object = {'source': source_name, 'field': field, 'value': value, 'configuration': configuration_name}
                checkDoc = FieldColl.find_one({'hash': Deduplication.generate_hash_from_obj(T2Object)})
                if checkDoc:
                    # we found a 100% matching T2 object
                    ioc.getLogger().trace("T2 object Located", trace=True)
                    update_statement = {
                        "$set": {
                            'score': int(checkDoc['score']) + 1,
                            'expiry': Deduplication.generate_expiry_time(expiry)
                        }
                    }
                    FieldColl.update_one(
                        {'_id': checkDoc['_id']},
                        update_statement
                    )
                    field_scores.append(100)
                    continue
                else:
                    # We have a possible unique value
                    ioc.getLogger().trace("T2 object not found", trace=True)
                    # generate a list to collect similarities to other field objects
                    fieldProbabilityList = []
                    for record in FieldColl.find({'source': source_name, 'configuration': configuration_name, 'field': field, 'type': 2})\
                            .sort('score', pymongo.ASCENDING).limit(100):
                        if Deduplication.string_match_percentage(record['value'], str(T2Object['value'])) > .95:
                            # We've found a REALLY strong match
                            # Set this field's score to that of the match
                            field_scores.append(
                                100 * Deduplication.string_match_percentage(record['value'], str(T2Object['value']))
                            )
                            # leave the for loop for this field since we found a highly probable match
                            break
                        else:
                            fieldProbabilityList.append(
                                100 * Deduplication.string_match_percentage(record['value'], str(T2Object['value']))
                            )
                    if fieldProbabilityList:
                        # We have at least one result
                        score = float(sum(fieldProbabilityList) / len(fieldProbabilityList))
                        ioc.getLogger().trace("Field Score [{0}]".format(score), verbose=True)
                        field_scores.append(score)
                    else:
                        # It is a globally unique field
                        field_scores.append(0)
                    # finally persist the new object
                    T2Object['hash'] = Deduplication.generate_hash_from_obj(T2Object)
                    T2Object['score'] = 1
                    T2Object['expiry'] = Deduplication.generate_expiry_time(expiry)
                    T2Object['max_expiry'] = Deduplication.generate_max_expiry_time(max_expiry)
                    T2Object['type'] = 2
                    T2Object['parentId'] = ObjectId(objectId)
                    FieldColl.insert_one(T2Object)
            else:
                ioc.getLogger().warning("field [{0}] not found in object".format(field), trace=True, notify=False)
                continue
        if len(field_scores) is 0:
            return 0.0
        else:
            return float(sum(field_scores) / float(len(field_scores)))

    @staticmethod
    def generate_hash_from_obj(obj):
        """Takes an object and generates a SHA256 Hash of it

        Args:
            obj (object): Hashable object ot generate a SHA256

        Returns:
            str: Object Hash

        """
        return hashlib.sha256(str(obj).encode('utf-8')).hexdigest()

    @staticmethod
    def generate_expiry_time(hours):
        """Generates UTC Timestamp for hours in the future

        Args:
            hours (int): How many hours in the future to expire on

        Returns:
            datetime.datetime: Datetime object for hours in the future

        """
        return datetime.datetime.utcnow() + datetime.timedelta(hours=int(hours))

    @staticmethod
    def generate_max_expiry_time(days):
        """Generates UTC Timestamp for hours in the future

        Args:
            days (int): How many days in the future to expire on

        Returns:
            datetime.datetime: Datetime object for days in the future

        """
        return datetime.datetime.utcnow() + datetime.timedelta(days=int(days))

    @staticmethod
    def string_match_percentage(constant, new_value):
        """Returns the percentage likelihood two strings are identical

        Args:
            constant (str): Value to use as base standard
            new_value (str): Value to compare `constant` against

        Returns:
            float: Percentage likelihood of duplicate value

        """
        return difflib.SequenceMatcher(lambda x: x == " ", constant, new_value).quick_ratio()



from tgt_grease.core.Types import Command
from tgt_grease.core import ImportTool
import importlib


class Help(Command):
    """The Help Command for GREASE

    Meant to provide a rich CLI Experience to users to enable quick help

    """

    purpose = "Provide Help Information"
    help = """
    Provide help information to users of GREASE about available commands. This though
    is just a demo of what you could print. Really it could be anything I suppose!
    
    Args:
        None
    """
    __author__ = "James E. Bell Jr."
    __version__ = "1.0.0"

    def __init__(self):
        super(Help, self).__init__()

    def execute(self, context):
        print("")
        print("Welcome to GREASE Help")
        impTool = ImportTool(self.ioc.getLogger())
        for route in self.ioc.getConfig().get('Import', 'searchPath'):
            mod = importlib.import_module(route)
            for attr in dir(mod):
                cmd = impTool.load(attr)
                if cmd and isinstance(cmd, Command):
                    print("<======================>")
                    print("[{0}] Purpose: [{1}]".format(
                        cmd.__class__.__name__,
                        cmd.purpose
                    ))
                    print("Author: {0}".format(cmd.__author__))
                    print("Current Version: {0}".format(cmd.__version__))
                    if cmd.os_needed:
                        print('Needs OS: {0}'.format(cmd.os_needed))
                    print(cmd.help)
                    print("<======================>")
        return True
